# Copyright (c) 2014 Ahmed H. Ismail
# Copyright (c) 2023 spdx contributors
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from beartype.typing import Any, Dict, List
from license_expression import get_spdx_licensing
from ply import yacc
from ply.yacc import LRParser

from spdx_tools.spdx.datetime_conversions import datetime_from_str
from spdx_tools.spdx.model import (
    Annotation,
    AnnotationType,
    CreationInfo,
    Document,
    ExternalDocumentRef,
    ExternalPackageRef,
    ExternalPackageRefCategory,
    ExtractedLicensingInfo,
    File,
    FileType,
    Package,
    PackagePurpose,
    PackageVerificationCode,
    Relationship,
    RelationshipType,
    Snippet,
    SpdxNoAssertion,
    SpdxNone,
    Version,
)
from spdx_tools.spdx.parser.actor_parser import ActorParser
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import (
    construct_or_raise_parsing_error,
    raise_parsing_error_if_logger_has_messages,
)
from spdx_tools.spdx.parser.tagvalue.helper_methods import (
    TAG_DATA_MODEL_FIELD,
    grammar_rule,
    parse_checksum,
    set_value,
    str_from_text,
)
from spdx_tools.spdx.parser.tagvalue.lexer import SPDXLexer

CLASS_MAPPING = dict(
    File="files",
    Annotation="annotations",
    Relationship="relationships",
    Snippet="snippets",
    Package="packages",
    ExtractedLicensingInfo="extracted_licensing_info",
)
ELEMENT_EXPECTED_START_TAG = dict(
    File="FileName",
    Annotation="Annotator",
    Relationship="Relationship",
    Snippet="SnippetSPDXID",
    Package="PackageName",
    ExtractedLicensingInfo="LicenseID",
)


class Parser:
    tokens: List[str]
    logger: Logger
    current_element: Dict[str, Any]
    creation_info: Dict[str, Any]
    elements_built: Dict[str, Any]
    lex: SPDXLexer
    yacc: LRParser

    def __init__(self, **kwargs):
        self.tokens = SPDXLexer.tokens
        self.logger = Logger()
        self.current_element = {"logger": Logger()}
        self.creation_info = {"logger": Logger()}
        self.elements_built = dict()
        self.lex = SPDXLexer()
        self.lex.build(reflags=re.UNICODE)
        self.yacc = yacc.yacc(module=self, **kwargs)

    @grammar_rule("start : start attrib ")
    def p_start_start_attrib(self, p):
        pass

    @grammar_rule("start : attrib ")
    def p_start_attrib(self, p):
        pass

    @grammar_rule(
        "attrib : spdx_version\n| spdx_id\n| data_license\n| doc_name\n| document_comment\n| document_namespace\n| "
        "creator\n| created\n| creator_comment\n| license_list_version\n| ext_doc_ref\n"
        # attributes for file
        "| file_name\n| file_type\n| file_checksum\n| file_license_concluded\n| file_license_info\n"
        "| file_copyright_text\n| file_license_comment\n| file_attribution_text\n| file_notice\n| file_comment\n"
        "| file_contributor\n"
        # attributes for annotation
        "| annotator\n| annotation_date\n| annotation_comment\n| annotation_type\n| annotation_spdx_id\n"
        # attributes for relationship
        "| relationship\n"
        # attributes for snippet
        "| snippet_spdx_id\n| snippet_name\n| snippet_comment\n| snippet_attribution_text\n| snippet_copyright_text\n"
        "| snippet_license_comment\n| file_spdx_id\n| snippet_license_concluded\n| snippet_license_info\n"
        "| snippet_byte_range\n| snippet_line_range\n"
        # attributes for package
        "| package_name\n| package_version\n| download_location\n| files_analyzed\n| homepage\n"
        "| summary\n| source_info\n| pkg_file_name\n| supplier\n| originator\n| pkg_checksum\n"
        "| verification_code\n| description\n| pkg_comment\n| pkg_attribution_text\n| pkg_license_declared\n"
        "| pkg_license_concluded\n| pkg_license_info\n| pkg_license_comment\n| pkg_copyright_text\n"
        "| pkg_external_ref\n| primary_package_purpose\n| built_date\n| release_date\n| valid_until_date\n"
        # attributes for extracted licensing info
        "| license_id\n| extracted_text\n| license_name\n| license_cross_ref\n| lic_comment\n"
        "| unknown_tag "
    )
    def p_attrib(self, p):
        pass

    # general parsing methods
    @grammar_rule(
        "license_id : LICENSE_ID error\n license_cross_ref : LICENSE_CROSS_REF error\n "
        "lic_comment : LICENSE_COMMENT error\n license_name : LICENSE_NAME error\n "
        "extracted_text : LICENSE_TEXT error\n "
        "file_name : FILE_NAME error\n file_contributor : FILE_CONTRIBUTOR error\n "
        "file_notice : FILE_NOTICE error\n file_copyright_text : FILE_COPYRIGHT_TEXT error\n "
        "file_license_comment : FILE_LICENSE_COMMENT error\n "
        "file_license_info : FILE_LICENSE_INFO error\n file_comment : FILE_COMMENT error\n "
        "file_checksum : FILE_CHECKSUM error\n file_license_concluded : FILE_LICENSE_CONCLUDED error\n "
        "file_type : FILE_TYPE error\n file_attribution_text : FILE_ATTRIBUTION_TEXT error\n "
        "package_name : PKG_NAME error\n pkg_attribution_text : PKG_ATTRIBUTION_TEXT error\n "
        "description : PKG_DESCRIPTION error\n pkg_comment : PKG_COMMENT error\n "
        "summary : PKG_SUMMARY error\n pkg_copyright_text : PKG_COPYRIGHT_TEXT error\n "
        "pkg_external_ref : PKG_EXTERNAL_REF error\n pkg_license_comment : PKG_LICENSE_COMMENT error\n "
        "pkg_license_declared : PKG_LICENSE_DECLARED error\n pkg_license_info : PKG_LICENSE_INFO error \n "
        "pkg_license_concluded : PKG_LICENSE_CONCLUDED error\n source_info : PKG_SOURCE_INFO error\n "
        "homepage : PKG_HOMEPAGE error\n pkg_checksum : PKG_CHECKSUM error\n "
        "verification_code : PKG_VERIFICATION_CODE error\n originator : PKG_ORIGINATOR error\n "
        "download_location : PKG_DOWNLOAD_LOCATION error\n files_analyzed : PKG_FILES_ANALYZED error\n "
        "supplier : PKG_SUPPLIER error\n pkg_file_name : PKG_FILE_NAME error\n "
        "package_version : PKG_VERSION error\n primary_package_purpose : PRIMARY_PACKAGE_PURPOSE error\n "
        "built_date : BUILT_DATE error\n release_date : RELEASE_DATE error\n "
        "valid_until_date : VALID_UNTIL_DATE error\n snippet_spdx_id : SNIPPET_SPDX_ID error\n "
        "snippet_name : SNIPPET_NAME error\n snippet_comment : SNIPPET_COMMENT error\n "
        "snippet_attribution_text : SNIPPET_ATTRIBUTION_TEXT error\n "
        "snippet_copyright_text : SNIPPET_COPYRIGHT_TEXT error\n "
        "snippet_license_comment : SNIPPET_LICENSE_COMMENT error\n file_spdx_id : SNIPPET_FILE_SPDXID error\n "
        "snippet_license_concluded : SNIPPET_LICENSE_CONCLUDED error\n "
        "snippet_license_info : SNIPPET_LICENSE_INFO error\n "
        "snippet_byte_range : SNIPPET_BYTE_RANGE error\n snippet_line_range : SNIPPET_LINE_RANGE error\n "
        "annotator : ANNOTATOR error\n annotation_date : ANNOTATION_DATE error\n "
        "annotation_comment : ANNOTATION_COMMENT error\n annotation_type : ANNOTATION_TYPE error\n "
        "annotation_spdx_id : ANNOTATION_SPDX_ID error\n relationship : RELATIONSHIP error"
    )
    def p_current_element_error(self, p):
        if p[1] in ELEMENT_EXPECTED_START_TAG.values():
            self.initialize_new_current_element(TAG_DATA_MODEL_FIELD[p[1]][0])
        self.current_element["logger"].append(
            f"Error while parsing {p[1]}: Token did not match specified grammar rule. Line: {p.lineno(1)}"
        )

    @grammar_rule(
        "license_name : LICENSE_NAME line_or_no_assertion\n extracted_text : LICENSE_TEXT text_or_line\n "
        "lic_comment : LICENSE_COMMENT text_or_line\n license_id : LICENSE_ID LINE\n "
        "file_name : FILE_NAME LINE \n file_notice : FILE_NOTICE text_or_line\n "
        "file_copyright_text : FILE_COPYRIGHT_TEXT line_or_no_assertion_or_none\n "
        "file_license_comment : FILE_LICENSE_COMMENT text_or_line\n "
        "file_comment : FILE_COMMENT text_or_line\n "
        "file_license_concluded : FILE_LICENSE_CONCLUDED license_or_no_assertion_or_none\n "
        "package_name : PKG_NAME LINE\n description : PKG_DESCRIPTION text_or_line\n "
        "summary : PKG_SUMMARY text_or_line\n source_info : PKG_SOURCE_INFO text_or_line\n "
        "homepage : PKG_HOMEPAGE line_or_no_assertion_or_none\n "
        "download_location : PKG_DOWNLOAD_LOCATION line_or_no_assertion_or_none\n "
        "originator : PKG_ORIGINATOR actor_or_no_assertion\n supplier : PKG_SUPPLIER actor_or_no_assertion\n "
        "pkg_comment : PKG_COMMENT text_or_line\n "
        "pkg_copyright_text : PKG_COPYRIGHT_TEXT line_or_no_assertion_or_none\n "
        "pkg_license_declared : PKG_LICENSE_DECLARED license_or_no_assertion_or_none\n "
        "pkg_file_name : PKG_FILE_NAME LINE\n "
        "pkg_license_concluded : PKG_LICENSE_CONCLUDED license_or_no_assertion_or_none\n "
        "package_version : PKG_VERSION LINE\n pkg_license_comment : PKG_LICENSE_COMMENT text_or_line\n "
        "snippet_spdx_id : SNIPPET_SPDX_ID LINE\n snippet_name : SNIPPET_NAME LINE\n "
        "snippet_comment : SNIPPET_COMMENT text_or_line\n "
        "snippet_copyright_text : SNIPPET_COPYRIGHT_TEXT line_or_no_assertion_or_none\n "
        "snippet_license_comment : SNIPPET_LICENSE_COMMENT text_or_line\n "
        "file_spdx_id : SNIPPET_FILE_SPDXID LINE\n "
        "snippet_license_concluded : SNIPPET_LICENSE_CONCLUDED license_or_no_assertion_or_none\n "
        "annotation_spdx_id : ANNOTATION_SPDX_ID LINE\n "
        "annotation_comment : ANNOTATION_COMMENT text_or_line"
    )
    def p_generic_value(self, p):
        if p[1] in ELEMENT_EXPECTED_START_TAG.values():
            self.initialize_new_current_element(TAG_DATA_MODEL_FIELD[p[1]][0])
        if self.check_that_current_element_matches_class_for_value(TAG_DATA_MODEL_FIELD[p[1]][0], p.lineno(1)):
            set_value(p, self.current_element)

    @grammar_rule(
        "unknown_tag : UNKNOWN_TAG text_or_line\n | UNKNOWN_TAG ISO8601_DATE\n | UNKNOWN_TAG PERSON_VALUE \n"
        "| UNKNOWN_TAG"
    )
    def p_unknown_tag(self, p):
        self.logger.append(f"Unknown tag provided in line {p.lineno(1)}")

    @grammar_rule("text_or_line : TEXT")
    def p_text(self, p):
        p[0] = str_from_text(p[1])

    @grammar_rule("text_or_line : LINE\n line_or_no_assertion : LINE\nline_or_no_assertion_or_none : text_or_line")
    def p_line(self, p):
        p[0] = p[1]

    @grammar_rule(
        "license_or_no_assertion_or_none : NO_ASSERTION\n actor_or_no_assertion : NO_ASSERTION\n"
        "line_or_no_assertion : NO_ASSERTION\n line_or_no_assertion_or_none : NO_ASSERTION"
    )
    def p_no_assertion(self, p):
        p[0] = SpdxNoAssertion()

    @grammar_rule("license_or_no_assertion_or_none : NONE\n line_or_no_assertion_or_none : NONE")
    def p_none(self, p):
        p[0] = SpdxNone()

    @grammar_rule("license_or_no_assertion_or_none : LINE")
    def p_license(self, p):
        p[0] = get_spdx_licensing().parse(p[1])

    @grammar_rule("actor_or_no_assertion : PERSON_VALUE\n | ORGANIZATION_VALUE")
    def p_actor_values(self, p):
        p[0] = ActorParser.parse_actor(p[1])

    @grammar_rule("spdx_id : SPDX_ID LINE")
    def p_spdx_id(self, p):
        # As all SPDX Ids share the same tag, there is no knowing which spdx_id belongs to the document.
        # We assume that to be the first spdx_id we encounter. As the specification does not explicitly require this,
        # our approach might lead to unwanted behavior when the document's SPDX Id is defined later in the document.
        if "spdx_id" in self.creation_info:
            self.current_element["spdx_id"] = p[2]
        else:
            self.creation_info["spdx_id"] = p[2]

    # parsing methods for creation info / document level

    @grammar_rule(
        "license_list_version : LICENSE_LIST_VERSION error\n document_comment : DOC_COMMENT error\n "
        "document_namespace : DOC_NAMESPACE error\n data_license : DOC_LICENSE error\n "
        "doc_name : DOC_NAME error\n ext_doc_ref : EXT_DOC_REF error\n spdx_version : DOC_VERSION error\n "
        "creator_comment : CREATOR_COMMENT error\n creator : CREATOR error\n created : CREATED error"
    )
    def p_creation_info_value_error(self, p):
        self.creation_info["logger"].append(
            f"Error while parsing {p[1]}: Token did not match specified grammar rule. Line: {p.lineno(1)}"
        )

    @grammar_rule(
        "document_comment : DOC_COMMENT text_or_line\n document_namespace : DOC_NAMESPACE LINE\n "
        "data_license : DOC_LICENSE LINE\n spdx_version : DOC_VERSION LINE\n "
        "creator_comment : CREATOR_COMMENT text_or_line\n doc_name : DOC_NAME LINE"
    )
    def p_generic_value_creation_info(self, p):
        set_value(p, self.creation_info)

    @grammar_rule("license_list_version : LICENSE_LIST_VERSION LINE")
    def p_license_list_version(self, p):
        set_value(p, self.creation_info, method_to_apply=Version.from_string)

    @grammar_rule("ext_doc_ref : EXT_DOC_REF LINE")
    def p_external_document_ref(self, p):
        external_doc_ref_regex = re.compile(r"(.*)(\s*SHA1:\s*[a-f0-9]{40})")
        external_doc_ref_match = external_doc_ref_regex.match(p[2])
        if not external_doc_ref_match:
            self.creation_info["logger"].append(
                f"Error while parsing ExternalDocumentRef: Couldn't match Checksum. Line: {p.lineno(1)}"
            )
            return
        try:
            document_ref_id, document_uri = external_doc_ref_match.group(1).strip().split(" ")
        except ValueError:
            self.creation_info["logger"].append(
                f"Error while parsing ExternalDocumentRef: Couldn't split the first part of the value into "
                f"document_ref_id and document_uri. Line: {p.lineno(1)}"
            )
            return
        checksum = parse_checksum(external_doc_ref_match.group(2).strip())
        external_document_ref = ExternalDocumentRef(document_ref_id, document_uri, checksum)
        self.creation_info.setdefault("external_document_refs", []).append(external_document_ref)

    @grammar_rule("creator : CREATOR PERSON_VALUE\n| CREATOR TOOL_VALUE\n| CREATOR ORGANIZATION_VALUE")
    def p_creator(self, p):
        self.creation_info.setdefault("creators", []).append(ActorParser.parse_actor(p[2]))

    @grammar_rule("created : CREATED ISO8601_DATE")
    def p_created(self, p):
        set_value(p, self.creation_info, method_to_apply=datetime_from_str)

    # parsing methods for extracted licensing info

    @grammar_rule("license_cross_ref : LICENSE_CROSS_REF LINE")
    def p_extracted_cross_reference(self, p):
        if self.check_that_current_element_matches_class_for_value(ExtractedLicensingInfo, p.lineno(1)):
            self.current_element.setdefault("cross_references", []).append(p[2])

    # parsing methods for file

    @grammar_rule("file_contributor : FILE_CONTRIBUTOR LINE")
    def p_file_contributor(self, p):
        if self.check_that_current_element_matches_class_for_value(File, p.lineno(1)):
            self.current_element.setdefault("contributors", []).append(p[2])

    @grammar_rule("file_attribution_text : FILE_ATTRIBUTION_TEXT text_or_line")
    def p_file_attribution_text(self, p):
        if self.check_that_current_element_matches_class_for_value(File, p.lineno(1)):
            self.current_element.setdefault("attribution_texts", []).append(p[2])

    @grammar_rule("file_license_info : FILE_LICENSE_INFO license_or_no_assertion_or_none")
    def p_file_license_info(self, p):
        if self.check_that_current_element_matches_class_for_value(File, p.lineno(1)):
            self.current_element.setdefault("license_info_in_file", []).append(p[2])

    @grammar_rule("file_type : FILE_TYPE LINE")
    def p_file_type(self, p):
        if not self.check_that_current_element_matches_class_for_value(File, p.lineno(1)):
            return
        try:
            file_type = FileType[p[2].strip()]
        except KeyError:
            self.current_element["logger"].append(f"Invalid FileType: {p[2]}. Line {p.lineno(1)}")
            return
        self.current_element.setdefault("file_types", []).append(file_type)

    @grammar_rule("file_checksum : FILE_CHECKSUM CHECKSUM")
    def p_file_checksum(self, p):
        if not self.check_that_current_element_matches_class_for_value(File, p.lineno(1)):
            return
        checksum = parse_checksum(p[2])
        self.current_element.setdefault("checksums", []).append(checksum)

    # parsing methods for package

    @grammar_rule("pkg_attribution_text : PKG_ATTRIBUTION_TEXT text_or_line")
    def p_pkg_attribution_text(self, p):
        self.check_that_current_element_matches_class_for_value(Package, p.lineno(1))
        self.current_element.setdefault("attribution_texts", []).append(p[2])

    @grammar_rule(
        "pkg_external_ref : PKG_EXTERNAL_REF LINE PKG_EXTERNAL_REF_COMMENT text_or_line\n | PKG_EXTERNAL_REF LINE"
    )
    def p_pkg_external_refs(self, p):
        if not self.check_that_current_element_matches_class_for_value(Package, p.lineno(1)):
            return
        try:
            category, reference_type, locator = p[2].split(" ")
        except ValueError:
            self.current_element["logger"].append(
                f"Couldn't split PackageExternalRef in category, reference_type and locator. Line: {p.lineno(1)}"
            )
            return
        comment = None
        if len(p) == 5:
            comment = p[4]
        try:
            category = ExternalPackageRefCategory[category.replace("-", "_")]
        except KeyError:
            self.current_element["logger"].append(
                f"Invalid ExternalPackageRefCategory: {category}. Line: {p.lineno(1)}"
            )
            return
        try:
            external_package_ref = construct_or_raise_parsing_error(
                ExternalPackageRef,
                {"category": category, "reference_type": reference_type, "locator": locator, "comment": comment},
            )
        except SPDXParsingError as err:
            self.current_element["logger"].append(err.get_messages())
            return
        self.current_element.setdefault("external_references", []).append(external_package_ref)

    @grammar_rule("pkg_license_info : PKG_LICENSE_INFO license_or_no_assertion_or_none")
    def p_pkg_license_info_from_file(self, p):
        if self.check_that_current_element_matches_class_for_value(Package, p.lineno(1)):
            self.current_element.setdefault("license_info_from_files", []).append(p[2])

    @grammar_rule("pkg_checksum : PKG_CHECKSUM CHECKSUM")
    def p_pkg_checksum(self, p):
        if not self.check_that_current_element_matches_class_for_value(Package, p.lineno(1)):
            return
        checksum = parse_checksum(p[2])
        self.current_element.setdefault("checksums", []).append(checksum)

    @grammar_rule("verification_code : PKG_VERIFICATION_CODE LINE")
    def p_pkg_verification_code(self, p):
        if not self.check_that_current_element_matches_class_for_value(Package, p.lineno(1)):
            return

        if "verification_code" in self.current_element:
            self.current_element["logger"].append(f"Multiple values for {p[1]} found. Line: {p.lineno(1)}")
            return
        verif_code_regex = re.compile(r"([0-9a-f]{40})\s*(\(excludes:\s*(.+)\))?", re.UNICODE)
        verif_code_code_grp = 1
        verif_code_exc_files_grp = 3
        match = verif_code_regex.match(p[2])
        if not match:
            self.current_element["logger"].append(
                f"Error while parsing {p[1]}: Value did not match expected format. Line: {p.lineno(1)}"
            )
            return
        value = match.group(verif_code_code_grp)
        excluded_files = None
        if match.group(verif_code_exc_files_grp):
            excluded_files = match.group(verif_code_exc_files_grp).split(",")
        self.current_element["verification_code"] = PackageVerificationCode(value, excluded_files)

    @grammar_rule("files_analyzed : PKG_FILES_ANALYZED LINE")
    def p_pkg_files_analyzed(self, p):
        if not self.check_that_current_element_matches_class_for_value(Package, p.lineno(1)):
            return
        if "files_analyzed" in self.current_element:
            self.current_element["logger"].append(f"Multiple values for {p[1]} found. Line: {p.lineno(1)}")
            return
        self.current_element["files_analyzed"] = p[2] in ["true", "True"]

    @grammar_rule("primary_package_purpose : PRIMARY_PACKAGE_PURPOSE LINE")
    def p_primary_package_purpose(self, p):
        if self.check_that_current_element_matches_class_for_value(Package, p.lineno(1)):
            set_value(p, self.current_element, method_to_apply=lambda x: PackagePurpose[x.replace("-", "_")])

    @grammar_rule(
        "built_date : BUILT_DATE ISO8601_DATE\n release_date : RELEASE_DATE ISO8601_DATE\n "
        "valid_until_date : VALID_UNTIL_DATE ISO8601_DATE"
    )
    def p_package_dates(self, p):
        if self.check_that_current_element_matches_class_for_value(Package, p.lineno(1)):
            set_value(p, self.current_element, method_to_apply=datetime_from_str)

    # parsing methods for snippet

    @grammar_rule("snippet_attribution_text : SNIPPET_ATTRIBUTION_TEXT text_or_line")
    def p_snippet_attribution_text(self, p):
        if self.check_that_current_element_matches_class_for_value(Snippet, p.lineno(1)):
            self.current_element.setdefault("attribution_texts", []).append(p[2])

    @grammar_rule("snippet_license_info : SNIPPET_LICENSE_INFO license_or_no_assertion_or_none")
    def p_snippet_license_info(self, p):
        if self.check_that_current_element_matches_class_for_value(Snippet, p.lineno(1)):
            self.current_element.setdefault("license_info_in_snippet", []).append(p[2])

    @grammar_rule("snippet_byte_range : SNIPPET_BYTE_RANGE LINE\n snippet_line_range : SNIPPET_LINE_RANGE LINE")
    def p_snippet_range(self, p):
        if not self.check_that_current_element_matches_class_for_value(Snippet, p.lineno(1)):
            return

        argument_name = TAG_DATA_MODEL_FIELD[p[1]][1]
        if argument_name in self.current_element:
            self.current_element["logger"].append(f"Multiple values for {p[1]} found. Line: {p.lineno(1)}")
            return
        range_re = re.compile(r"^(\d+):(\d+)$", re.UNICODE)
        if not range_re.match(p[2].strip()):
            self.current_element["logger"].append(
                f"Value for {p[1]} doesn't match valid range pattern. " f"Line: {p.lineno(1)}"
            )
            return
        startpoint = int(p[2].split(":")[0])
        endpoint = int(p[2].split(":")[-1])
        self.current_element[argument_name] = startpoint, endpoint

    # parsing methods for annotation

    @grammar_rule("annotator : ANNOTATOR PERSON_VALUE\n| TOOL_VALUE\n| ORGANIZATION_VALUE")
    def p_annotator(self, p):
        self.initialize_new_current_element(Annotation)
        set_value(p, self.current_element, method_to_apply=ActorParser.parse_actor)

    @grammar_rule("annotation_date : ANNOTATION_DATE ISO8601_DATE")
    def p_annotation_date(self, p):
        if self.check_that_current_element_matches_class_for_value(Annotation, p.lineno(1)):
            set_value(p, self.current_element, method_to_apply=datetime_from_str)

    @grammar_rule("annotation_type : ANNOTATION_TYPE LINE")
    def p_annotation_type(self, p):
        if self.check_that_current_element_matches_class_for_value(Annotation, p.lineno(1)):
            set_value(p, self.current_element, method_to_apply=lambda x: AnnotationType[x])

    # parsing methods for relationship

    @grammar_rule("relationship : RELATIONSHIP LINE RELATIONSHIP_COMMENT text_or_line\n " "| RELATIONSHIP LINE")
    def p_relationship(self, p):
        self.initialize_new_current_element(Relationship)
        try:
            spdx_element_id, relationship_type, related_spdx_element_id = p[2].split(" ")
        except ValueError:
            self.current_element["logger"].append(
                f"Relationship couldn't be split in spdx_element_id, relationship_type and "
                f"related_spdx_element. Line: {p.lineno(1)}"
            )
            return
        try:
            self.current_element["relationship_type"] = RelationshipType[relationship_type]
        except KeyError:
            self.current_element["logger"].append(f"Invalid RelationshipType {relationship_type}. Line: {p.lineno(1)}")
        if related_spdx_element_id == "NONE":
            related_spdx_element_id = SpdxNone()
        if related_spdx_element_id == "NOASSERTION":
            related_spdx_element_id = SpdxNoAssertion()
        self.current_element["related_spdx_element_id"] = related_spdx_element_id
        self.current_element["spdx_element_id"] = spdx_element_id
        if len(p) == 5:
            self.current_element["comment"] = p[4]

    def p_error(self, p):
        pass

    def parse(self, text):
        # entry point for the tag-value parser
        self.yacc.parse(text, lexer=self.lex)
        # this constructs the last remaining element; all other elements are constructed at the start of
        # their subsequent element
        self.construct_current_element()

        # To be able to parse creation info values if they appear in between other elements, e.g. packages, we use
        # two different dictionaries to collect the creation info and all other elements. Therefore, we have a separate
        # logger for the creation info whose messages we need to add to the main logger to than raise all collected
        # messages at once.
        creation_info_logger = self.creation_info.pop("logger")
        if creation_info_logger.has_messages():
            self.logger.extend([f"Error while parsing CreationInfo: {creation_info_logger.get_messages()}"])

        raise_parsing_error_if_logger_has_messages(self.logger)
        creation_info = construct_or_raise_parsing_error(CreationInfo, self.creation_info)
        self.elements_built["creation_info"] = creation_info
        document = construct_or_raise_parsing_error(Document, self.elements_built)
        return document

    def initialize_new_current_element(self, clazz: Any):
        self.construct_current_element()
        self.current_element["class"] = clazz

    def check_that_current_element_matches_class_for_value(self, expected_class, line_number) -> bool:
        if "class" not in self.current_element or expected_class != self.current_element["class"]:
            self.logger.append(
                f"Element {expected_class.__name__} is not the current element in scope, probably the expected tag to "
                f"start the element ({ELEMENT_EXPECTED_START_TAG[expected_class.__name__]}) is missing. "
                f"Line: {line_number}"
            )
            return False
        return True

    def construct_current_element(self):
        if "class" not in self.current_element:
            # This happens when the first element is initialized via initialize_new_current_element() or if the first
            # element is missing its expected starting tag. In both cases we are unable to construct an element.
            return

        clazz = self.current_element.pop("class")
        try:
            raise_parsing_error_if_logger_has_messages(self.current_element.pop("logger"), clazz.__name__)
            self.elements_built.setdefault(CLASS_MAPPING[clazz.__name__], []).append(
                construct_or_raise_parsing_error(clazz, self.current_element)
            )
            if clazz == File:
                self.check_for_preceding_package_and_build_contains_relationship()
        except SPDXParsingError as err:
            self.logger.extend(err.get_messages())
        self.current_element = {"logger": Logger()}

    def check_for_preceding_package_and_build_contains_relationship(self):
        file_spdx_id = self.current_element["spdx_id"]
        if "packages" not in self.elements_built:
            return
        # We assume that all files that are not contained in a package precede any package information. Any file
        # information that follows any package information is assigned to the last parsed package by creating a
        # corresponding contains relationship.
        # (see https://spdx.github.io/spdx-spec/v2.3/composition-of-an-SPDX-document/#5.2.2)
        if not self.elements_built["packages"]:
            self.logger.append(
                f"Error while building contains relationship for file {file_spdx_id}, "
                f"preceding package was not parsed successfully."
            )
            return
        package_spdx_id = self.elements_built["packages"][-1].spdx_id
        relationship = Relationship(package_spdx_id, RelationshipType.CONTAINS, file_spdx_id)
        if relationship not in self.elements_built.setdefault("relationships", []):
            self.elements_built["relationships"].append(relationship)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values


@dataclass_with_properties
class NamespaceMap:
    prefix: str
    namespace: str  # anyURI

    def __init__(self, prefix: str, namespace: str):
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import auto

from spdx_tools.spdx.jsonschema.json_property import JsonProperty


class ExtractedLicensingInfoProperty(JsonProperty):
    COMMENT = auto()
    EXTRACTED_TEXT = auto()
    LICENSE_ID = auto()
    NAME = auto()
    SEE_ALSOS = auto()



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from unittest import mock

import pytest

from spdx_tools.spdx.model import ExternalDocumentRef


@mock.patch("spdx_tools.spdx.model.Checksum", autospec=True)
def test_correct_initialization(checksum):
    external_document_ref = ExternalDocumentRef("id", "uri", checksum)
    assert external_document_ref.document_ref_id == "id"
    assert external_document_ref.document_uri == "uri"
    assert external_document_ref.checksum == checksum


@mock.patch("spdx_tools.spdx.model.Checksum", autospec=True)
def test_wrong_type_in_spdx_id(checksum):
    with pytest.raises(TypeError):
        ExternalDocumentRef(42, "uri", checksum)


@mock.patch("spdx_tools.spdx.model.Checksum", autospec=True)
def test_wrong_type_in_document_uri(checksum):
    with pytest.raises(TypeError):
        ExternalDocumentRef("id", 42, checksum)


def test_wrong_type_in_checksum():
    with pytest.raises(TypeError):
        ExternalDocumentRef("id", "uri", 42)



#  SPDX-FileCopyrightText: 2023 spdx contributors
#
#  SPDX-License-Identifier: Apache-2.0
import logging
from datetime import datetime
from typing import List

from license_expression import get_spdx_licensing

from spdx_tools.spdx.model import (
    Actor,
    ActorType,
    Checksum,
    ChecksumAlgorithm,
    CreationInfo,
    Document,
    ExternalPackageRef,
    ExternalPackageRefCategory,
    File,
    FileType,
    Package,
    PackagePurpose,
    PackageVerificationCode,
    Relationship,
    RelationshipType,
)
from spdx_tools.spdx.validation.document_validator import validate_full_spdx_document
from spdx_tools.spdx.validation.validation_message import ValidationMessage

# This example shows how to use the spdx-tools to create an SPDX document from scratch,
# validate it and write it to a file.


def test_spdx2_document_from_scratch():
    # First up, we need general information about the creation of the document, summarised by the CreationInfo class.
    creation_info = CreationInfo(
        spdx_version="SPDX-2.3",
        spdx_id="SPDXRef-DOCUMENT",
        name="document name",
        data_license="CC0-1.0",
        document_namespace="https://some.namespace",
        creators=[Actor(ActorType.PERSON, "Jane Doe", "jane.doe@example.com")],
        created=datetime(2022, 1, 1),
    )

    # creation_info is the only required property of the Document class (have a look there!),
    # the rest are optional lists.
    # So, we are set up to create a new document instance.
    document = Document(creation_info)

    # The document currently does not describe anything. Let's create a package that we can add to it.
    # The Package class has quite a few properties (have a look there!),
    # but only name, spdx_id and download_location are mandatory in SPDX v2.3.
    package = Package(
        name="package name",
        spdx_id="SPDXRef-Package",
        download_location="https://download.com",
        version="2.2.1",
        file_name="./foo.bar",
        supplier=Actor(ActorType.PERSON, "Jane Doe", "jane.doe@example.com"),
        originator=Actor(ActorType.ORGANIZATION, "some organization", "contact@example.com"),
        files_analyzed=True,
        verification_code=PackageVerificationCode(
            value="d6a770ba38583ed4bb4525bd96e50461655d2758", excluded_files=["./some.file"]
        ),
        checksums=[
            Checksum(ChecksumAlgorithm.SHA1, "d6a770ba38583ed4bb4525bd96e50461655d2758"),
            Checksum(ChecksumAlgorithm.MD5, "624c1abb3664f4b35547e7c73864ad24"),
        ],
        license_concluded=get_spdx_licensing().parse("GPL-2.0-only OR MIT"),
        license_info_from_files=[get_spdx_licensing().parse("GPL-2.0-only"), get_spdx_licensing().parse("MIT")],
        license_declared=get_spdx_licensing().parse("GPL-2.0-only AND MIT"),
        license_comment="license comment",
        copyright_text="Copyright 2022 Jane Doe",
        description="package description",
        attribution_texts=["package attribution"],
        primary_package_purpose=PackagePurpose.LIBRARY,
        release_date=datetime(2015, 1, 1),
        external_references=[
            ExternalPackageRef(
                category=ExternalPackageRefCategory.OTHER,
                reference_type="http://reference.type",
                locator="reference/locator",
                comment="external reference comment",
            )
        ],
    )

    # Now that we have a package defined, we can add it to the document's package property.
    document.packages = [package]

    # A DESCRIBES relationship asserts that the document indeed describes the package.
    describes_relationship = Relationship("SPDXRef-DOCUMENT", RelationshipType.DESCRIBES, "SPDXRef-Package")
    document.relationships = [describes_relationship]

    # Let's add two files. Have a look at the file class for all possible properties a file can have.
    file1 = File(
        name="./package/file1.py",
        spdx_id="SPDXRef-File1",
        file_types=[FileType.SOURCE],
        checksums=[
            Checksum(ChecksumAlgorithm.SHA1, "d6a770ba38583ed4bb4525bd96e50461655d2758"),
            Checksum(ChecksumAlgorithm.MD5, "624c1abb3664f4b35547e7c73864ad24"),
        ],
        license_concluded=get_spdx_licensing().parse("MIT"),
        license_info_in_file=[get_spdx_licensing().parse("MIT")],
        copyright_text="Copyright 2022 Jane Doe",
    )
    file2 = File(
        name="./package/file2.py",
        spdx_id="SPDXRef-File2",
        checksums=[
            Checksum(ChecksumAlgorithm.SHA1, "d6a770ba38583ed4bb4525bd96e50461655d2759"),
        ],
        license_concluded=get_spdx_licensing().parse("GPL-2.0-only"),
    )

    # Assuming the package contains those two files, we create two CONTAINS relationships.
    contains_relationship1 = Relationship("SPDXRef-Package", RelationshipType.CONTAINS, "SPDXRef-File1")
    contains_relationship2 = Relationship("SPDXRef-Package", RelationshipType.CONTAINS, "SPDXRef-File2")

    # This library uses run-time type checks when assigning properties.
    # Because in-place alterations like .append() circumvent these checks, we don't use them here.
    document.relationships += [contains_relationship1, contains_relationship2]
    document.files += [file1, file2]

    # We now have created a document with basic creation information, describing a package that contains two files.
    # You can also add Annotations, Snippets and ExtractedLicensingInfo
    # to the document in an analogous manner to the above.
    # Have a look at their respective classes if you are unsure about their properties.
    assert len(document.packages) == 1
    assert len(document.files) == 2
    assert len(document.relationships) == 3
    assert len(document.snippets) == 0
    assert len(document.annotations) == 0
    assert len(document.extracted_licensing_info) == 0

    # This library provides comprehensive validation against the SPDX specification.
    # Note that details of the validation depend on the SPDX version of the document.
    validation_messages: List[ValidationMessage] = validate_full_spdx_document(document)

    # You can have a look at each entry's message and context (like spdx_id, parent_id, full_element)
    # which will help you pinpoint the location of the invalidity.
    for message in validation_messages:
        logging.warning(message.validation_message)
        logging.warning(message.context)

    # If the document is valid, validation_messages will be empty.
    assert validation_messages == []



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Any, Type

from spdx_tools.spdx.jsonschema.converter import TypedConverter
from spdx_tools.spdx.jsonschema.extracted_licensing_info_properties import ExtractedLicensingInfoProperty
from spdx_tools.spdx.jsonschema.json_property import JsonProperty
from spdx_tools.spdx.jsonschema.optional_utils import apply_if_present
from spdx_tools.spdx.model import Document, ExtractedLicensingInfo


class ExtractedLicensingInfoConverter(TypedConverter[ExtractedLicensingInfo]):
    def _get_property_value(
        self,
        extracted_licensing_info: ExtractedLicensingInfo,
        extracted_licensing_info_property: ExtractedLicensingInfoProperty,
        document: Document = None,
    ) -> Any:
        if extracted_licensing_info_property == ExtractedLicensingInfoProperty.COMMENT:
            return extracted_licensing_info.comment
        elif extracted_licensing_info_property == ExtractedLicensingInfoProperty.EXTRACTED_TEXT:
            return extracted_licensing_info.extracted_text
        elif extracted_licensing_info_property == ExtractedLicensingInfoProperty.LICENSE_ID:
            return extracted_licensing_info.license_id
        elif extracted_licensing_info_property == ExtractedLicensingInfoProperty.NAME:
            return apply_if_present(str, extracted_licensing_info.license_name)
        elif extracted_licensing_info_property == ExtractedLicensingInfoProperty.SEE_ALSOS:
            return extracted_licensing_info.cross_references or None

    def get_json_type(self) -> Type[JsonProperty]:
        return ExtractedLicensingInfoProperty

    def get_data_model_type(self) -> Type[ExtractedLicensingInfo]:
        return ExtractedLicensingInfo



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import pytest

from spdx_tools.spdx.model import ExternalPackageRef, ExternalPackageRefCategory


def test_correct_initialization():
    external_package_reference = ExternalPackageRef(ExternalPackageRefCategory.OTHER, "type", "locator", "comment")
    assert external_package_reference.category == ExternalPackageRefCategory.OTHER
    assert external_package_reference.reference_type == "type"
    assert external_package_reference.locator == "locator"
    assert external_package_reference.comment == "comment"


def test_wrong_type_in_category():
    with pytest.raises(TypeError):
        ExternalPackageRef([ExternalPackageRefCategory.OTHER], "type", "locator")


def test_wrong_type_in_reference_type():
    with pytest.raises(TypeError):
        ExternalPackageRef(ExternalPackageRefCategory.OTHER, 42, "locator")


def test_wrong_type_in_locator():
    with pytest.raises(TypeError):
        ExternalPackageRef(ExternalPackageRefCategory.OTHER, "type", 42)


def test_wrong_type_in_comment():
    with pytest.raises(TypeError):
        ExternalPackageRef(ExternalPackageRefCategory.OTHER, "type", "locator", [])



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Callable, Optional, TypeVar

T = TypeVar("T")
S = TypeVar("S")


def apply_if_present(function: Callable[[T], S], optional_value: Optional[T]) -> Optional[S]:
    """
    Apply the passed function to the optional value if it is not None. Else returns None.
    """
    return function(optional_value) if optional_value is not None else None



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model import Relationship
from spdx_tools.spdx3.writer.console.console import write_value
from spdx_tools.spdx3.writer.console.element_writer import write_element_properties


def write_relationship(relationship: Relationship, text_output: TextIO, heading: bool = True):
    if heading:
        text_output.write("## Relationship\n")
    write_element_properties(relationship, text_output)
    for property_name in relationship.__annotations__.keys():
        write_value(property_name, getattr(relationship, property_name), text_output)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Any, Dict, Tuple, Type

from spdx_tools.spdx.jsonschema.annotation_converter import AnnotationConverter
from spdx_tools.spdx.jsonschema.converter import TypedConverter
from spdx_tools.spdx.jsonschema.json_property import JsonProperty
from spdx_tools.spdx.jsonschema.optional_utils import apply_if_present
from spdx_tools.spdx.jsonschema.snippet_properties import SnippetProperty
from spdx_tools.spdx.model import Document, Snippet


class SnippetConverter(TypedConverter[Snippet]):
    annotation_converter: AnnotationConverter

    def __init__(self):
        self.annotation_converter = AnnotationConverter()

    def json_property_name(self, snippet_property: SnippetProperty) -> str:
        if snippet_property == SnippetProperty.SPDX_ID:
            return "SPDXID"
        return super().json_property_name(snippet_property)

    def _get_property_value(
        self, snippet: Snippet, snippet_property: SnippetProperty, document: Document = None
    ) -> Any:
        if snippet_property == SnippetProperty.SPDX_ID:
            return snippet.spdx_id
        elif snippet_property == SnippetProperty.ANNOTATIONS:
            snippet_annotations = filter(
                lambda annotation: annotation.spdx_id == snippet.spdx_id, document.annotations
            )
            return [self.annotation_converter.convert(annotation) for annotation in snippet_annotations] or None
        elif snippet_property == SnippetProperty.ATTRIBUTION_TEXTS:
            return snippet.attribution_texts or None
        elif snippet_property == SnippetProperty.COMMENT:
            return snippet.comment
        elif snippet_property == SnippetProperty.COPYRIGHT_TEXT:
            return apply_if_present(str, snippet.copyright_text)
        elif snippet_property == SnippetProperty.LICENSE_COMMENTS:
            return snippet.license_comment
        elif snippet_property == SnippetProperty.LICENSE_CONCLUDED:
            return apply_if_present(str, snippet.license_concluded)
        elif snippet_property == SnippetProperty.LICENSE_INFO_IN_SNIPPETS:
            return [str(license_expression) for license_expression in snippet.license_info_in_snippet] or None
        elif snippet_property == SnippetProperty.NAME:
            return snippet.name
        elif snippet_property == SnippetProperty.RANGES:
            ranges = [convert_byte_range_to_dict(snippet.byte_range, snippet.file_spdx_id)]
            if snippet.line_range:
                ranges.append(convert_line_range_to_dict(snippet.line_range, snippet.file_spdx_id))
            return ranges
        elif snippet_property == SnippetProperty.SNIPPET_FROM_FILE:
            return snippet.file_spdx_id

    def get_json_type(self) -> Type[JsonProperty]:
        return SnippetProperty

    def get_data_model_type(self) -> Type[Snippet]:
        return Snippet

    def requires_full_document(self) -> bool:
        return True


def convert_line_range_to_dict(line_range: Tuple[int, int], file_id: str) -> Dict:
    return _convert_range_to_dict(line_range, file_id, "lineNumber")


def convert_byte_range_to_dict(byte_range: Tuple[int, int], file_id: str) -> Dict:
    return _convert_range_to_dict(byte_range, file_id, "offset")


def _convert_range_to_dict(int_range: Tuple[int, int], file_id: str, pointer_property: str) -> Dict:
    return {
        "startPointer": _pointer(file_id, int_range[0], pointer_property),
        "endPointer": _pointer(file_id, int_range[1], pointer_property),
    }


def _pointer(reference: str, target: int, pointer_property: str) -> Dict:
    return {"reference": reference, pointer_property: target}



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest.mock import NonCallableMagicMock


def assert_mock_method_called_with_arguments(mock_object: NonCallableMagicMock, method_name: str, *args):
    assert len(mock_object.method_calls) == len(args)
    for running_index in range(len(args)):
        call = mock_object.method_calls[running_index]
        assert call[0] == method_name
        assert call[1][0] == args[running_index]


def assert_no_mock_methods_called(mock_object: NonCallableMagicMock):
    assert len(mock_object.method_calls) == 0



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest import TestCase

import pytest

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.model import Relationship, RelationshipType, SpdxNoAssertion
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.jsonlikedict.relationship_parser import RelationshipParser


def test_parse_relationship():
    relationship_parser = RelationshipParser()

    relationship_dict = {
        "spdxElementId": DOCUMENT_SPDX_ID,
        "relationshipType": "CONTAINS",
        "relatedSpdxElement": "NOASSERTION",
        "comment": "Comment.",
    }

    relationship = relationship_parser.parse_relationship(relationship_dict)

    assert relationship.relationship_type == RelationshipType.CONTAINS
    assert relationship.spdx_element_id == DOCUMENT_SPDX_ID
    assert relationship.related_spdx_element_id == SpdxNoAssertion()
    assert relationship.comment == "Comment."


def test_parse_incomplete_relationship():
    relationship_parser = RelationshipParser()
    relationship_dict = {
        "spdxElementId": DOCUMENT_SPDX_ID,
        "relatedSpdxElement": "SPDXRef-Package",
        "comment": "Comment.",
    }

    with pytest.raises(SPDXParsingError):
        relationship_parser.parse_relationship(relationship_dict)


def test_parse_relationship_type():
    relationship_parser = RelationshipParser()
    relationship_type_str = "DEPENDENCY_OF"

    relationship_type = relationship_parser.parse_relationship_type(relationship_type_str)
    assert relationship_type == RelationshipType.DEPENDENCY_OF


def test_parse_document_describes():
    relationship_parser = RelationshipParser()

    document_dict = {
        "SPDXID": DOCUMENT_SPDX_ID,
        "documentDescribes": ["SPDXRef-Package", "SPDXRef-File", "SPDXRef-Snippet"],
    }

    relationships = relationship_parser.parse_document_describes(
        doc_spdx_id=DOCUMENT_SPDX_ID,
        described_spdx_ids=document_dict.get("documentDescribes"),
        existing_relationships=[],
    )

    assert len(relationships) == 3
    TestCase().assertCountEqual(
        relationships,
        [
            Relationship(DOCUMENT_SPDX_ID, RelationshipType.DESCRIBES, "SPDXRef-Package"),
            Relationship(DOCUMENT_SPDX_ID, RelationshipType.DESCRIBES, "SPDXRef-File"),
            Relationship(DOCUMENT_SPDX_ID, RelationshipType.DESCRIBES, "SPDXRef-Snippet"),
        ],
    )


@pytest.mark.parametrize(
    "document_describes,relationships,parsed_relationships",
    [
        (
            ["SPDXRef-Package", "SPDXRef-File"],
            [
                {
                    "spdxElementId": DOCUMENT_SPDX_ID,
                    "relatedSpdxElement": "SPDXRef-Package",
                    "relationshipType": "DESCRIBES",
                    "comment": "This relationship has a comment.",
                },
                {
                    "spdxElementId": "SPDXRef-File",
                    "relatedSpdxElement": DOCUMENT_SPDX_ID,
                    "relationshipType": "DESCRIBED_BY",
                    "comment": "This relationship has a comment.",
                },
            ],
            [
                Relationship(
                    related_spdx_element_id="SPDXRef-Package",
                    relationship_type=RelationshipType.DESCRIBES,
                    spdx_element_id=DOCUMENT_SPDX_ID,
                    comment="This relationship has a comment.",
                ),
                Relationship(
                    related_spdx_element_id=DOCUMENT_SPDX_ID,
                    relationship_type=RelationshipType.DESCRIBED_BY,
                    spdx_element_id="SPDXRef-File",
                    comment="This relationship has a comment.",
                ),
            ],
        ),
        (
            ["SPDXRef-Package", "SPDXRef-File", "SPDXRef-Package"],
            [],
            [
                Relationship(
                    related_spdx_element_id="SPDXRef-Package",
                    relationship_type=RelationshipType.DESCRIBES,
                    spdx_element_id=DOCUMENT_SPDX_ID,
                ),
                Relationship(
                    related_spdx_element_id="SPDXRef-File",
                    relationship_type=RelationshipType.DESCRIBES,
                    spdx_element_id=DOCUMENT_SPDX_ID,
                ),
            ],
        ),
    ],
)
def test_parse_document_describes_without_duplicating_relationships(
    document_describes, relationships, parsed_relationships
):
    relationship_parser = RelationshipParser()
    document_dict = {
        "SPDXID": DOCUMENT_SPDX_ID,
        "documentDescribes": document_describes,
        "relationships": relationships,
    }

    relationships = relationship_parser.parse_all_relationships(document_dict)

    assert len(relationships) == len(parsed_relationships)
    TestCase().assertCountEqual(relationships, parsed_relationships)


def test_parse_has_files():
    relationship_parser = RelationshipParser()
    document_dict = {"packages": [{"SPDXID": "SPDXRef-Package", "hasFiles": ["SPDXRef-File1", "SPDXRef-File2"]}]}

    relationships = relationship_parser.parse_has_files(document_dict.get("packages"), existing_relationships=[])

    assert len(relationships) == 2
    TestCase().assertCountEqual(
        relationships,
        [
            Relationship(
                spdx_element_id="SPDXRef-Package",
                relationship_type=RelationshipType.CONTAINS,
                related_spdx_element_id="SPDXRef-File1",
            ),
            Relationship(
                spdx_element_id="SPDXRef-Package",
                relationship_type=RelationshipType.CONTAINS,
                related_spdx_element_id="SPDXRef-File2",
            ),
        ],
    )


@pytest.mark.parametrize(
    "has_files,existing_relationships,contains_relationships",
    [
        (
            ["SPDXRef-File1", "SPDXRef-File2"],
            [
                Relationship(
                    spdx_element_id="SPDXRef-Package",
                    relationship_type=RelationshipType.CONTAINS,
                    related_spdx_element_id="SPDXRef-File1",
                    comment="This relationship has a comment.",
                ),
                Relationship(
                    spdx_element_id="SPDXRef-File2",
                    relationship_type=RelationshipType.CONTAINED_BY,
                    related_spdx_element_id="SPDXRef-Package",
                ),
            ],
            [],
        ),
        (
            ["SPDXRef-File1", "SPDXRef-File2", "SPDXRef-File1"],
            [],
            [
                Relationship(
                    spdx_element_id="SPDXRef-Package",
                    relationship_type=RelationshipType.CONTAINS,
                    related_spdx_element_id="SPDXRef-File1",
                ),
                Relationship(
                    spdx_element_id="SPDXRef-Package",
                    relationship_type=RelationshipType.CONTAINS,
                    related_spdx_element_id="SPDXRef-File2",
                ),
            ],
        ),
    ],
)
def test_parse_has_files_without_duplicating_relationships(has_files, existing_relationships, contains_relationships):
    relationship_parser = RelationshipParser()
    document_dict = {"packages": [{"SPDXID": "SPDXRef-Package", "hasFiles": has_files}]}
    relationships = relationship_parser.parse_has_files(
        document_dict.get("packages"), existing_relationships=existing_relationships
    )

    assert len(relationships) == len(contains_relationships)
    TestCase().assertCountEqual(relationships, contains_relationships)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import os

import pytest

from spdx_tools.spdx.model import Document
from spdx_tools.spdx.writer.rdf.rdf_writer import write_document_to_file
from tests.spdx.fixtures import document_fixture


@pytest.fixture
def temporary_file_path() -> str:
    temporary_file_path = "temp_test_rdf_writer_output.rdf.xml"
    yield temporary_file_path
    os.remove(temporary_file_path)


def test_write_document_to_file(temporary_file_path: str):
    document: Document = document_fixture()

    write_document_to_file(document, temporary_file_path, False)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import Enum

from beartype.typing import Any, Callable, Optional, Tuple, Type, Union
from rdflib import RDF, Graph, URIRef
from rdflib.exceptions import UniquenessError
from rdflib.namespace import NamespaceManager
from rdflib.term import BNode, Literal, Node

from spdx_tools.spdx.casing_tools import camel_case_to_snake_case
from spdx_tools.spdx.model import SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.model.spdx_no_assertion import SPDX_NO_ASSERTION_STRING
from spdx_tools.spdx.model.spdx_none import SPDX_NONE_STRING
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE


def parse_literal(
    logger: Logger,
    graph: Graph,
    subject: Node,
    predicate: Node,
    parsing_method: Callable = lambda x: x.strip(),
    default: Any = None,
):
    value = get_unique_value(logger, graph, subject, predicate, default)
    if not value:
        return default
    return apply_parsing_method_or_log_error(logger, value, parsing_method, default)


def apply_parsing_method_or_log_error(
    logger: Logger, value: Any, parsing_method: Callable = lambda x: x.strip(), default: Any = None
):
    try:
        return parsing_method(value)
    except SPDXParsingError as err:
        logger.extend(err.get_messages())
    except (TypeError, ValueError) as err:
        logger.append(err.args[0])
    return default


def parse_literal_or_no_assertion_or_none(
    logger: Logger,
    graph: Graph,
    subject: Node,
    predicate: Node,
    parsing_method: Callable = lambda x: x.strip(),
    default: Any = None,
):
    value = get_unique_value(logger, graph, subject, predicate, default)
    return get_correctly_typed_value(logger, value, parsing_method, default)


def get_correctly_typed_value(
    logger: Logger, value: Any, parsing_method: Callable = lambda x: x.strip(), default: Any = None
):
    if not value:
        return default
    if value == SPDX_NAMESPACE.noassertion or value.toPython() == SPDX_NO_ASSERTION_STRING:
        return SpdxNoAssertion()
    if value == SPDX_NAMESPACE.none or value.toPython() == SPDX_NONE_STRING:
        return SpdxNone()
    return apply_parsing_method_or_log_error(logger, value, parsing_method, default)


def get_unique_value(logger: Logger, graph: Graph, subject: Node, predicate: Node, default: Any) -> Any:
    try:
        value = graph.value(subject=subject, predicate=predicate, default=default, any=False)
        return value
    except UniquenessError:
        logger.append(f"Multiple values for unique value {predicate} found.")
        return default


def parse_enum_value(enum_str: str, enum_class: Type[Enum], prefix: str) -> Enum:
    try:
        enum_without_rdf_prefix = remove_prefix(enum_str, prefix)
        value = camel_case_to_snake_case(enum_without_rdf_prefix).upper()
        return enum_class[value]
    except KeyError:
        raise SPDXParsingError([f"Invalid value for {enum_class}: {enum_str}"])


def parse_spdx_id(resource: Union[URIRef, BNode], doc_namespace: str, graph: Graph) -> Optional[str]:
    if not resource or isinstance(resource, BNode):
        return None
    if resource.startswith(f"{doc_namespace}#"):
        return resource.fragment
    if "#" in resource:
        namespace_manager = NamespaceManager(graph)
        return namespace_manager.normalizeUri(resource)
    return resource.toPython() or None


# Python 3.9 introduced the method removeprefix() for strings, but as we are also supporting Python 3.7 and 3.8 we need
# to write our own helper method to delete prefixes.
def remove_prefix(string: str, prefix: str) -> str:
    if string.startswith(prefix):
        return string[len(prefix) :]
    return string


def get_correctly_typed_triples(
    logger: Logger,
    graph: Graph,
    subject: Optional[Node] = None,
    predicate: Optional[Node] = None,
    _object: Optional[Node] = None,
) -> Tuple[Union[BNode, URIRef], Node, Union[BNode, Literal, URIRef]]:
    # this is a helper method to cast some rdf types from graph.triples() to be compatible with the
    # code that follows
    for s, p, o in graph.triples((subject, predicate, _object)):
        if not isinstance(s, (BNode, URIRef)):
            logger.append(
                f"Warning: Subject {s} should be of type BNode or URIRef, but is {type(s).__name__}. "
                f"This might lead to a failure."
            )
        if not isinstance(o, (BNode, Literal, URIRef)):
            logger.append(
                f"Warning: Object {o} should be of type BNode, Literal or URIRef, but is {type(o).__name__}. "
                f"This might lead to a failure."
            )
        yield s, p, o


def get_value_from_graph(
    logger: Logger,
    graph: Graph,
    subject: Optional[Node] = None,
    predicate: Optional[Node] = RDF.value,
    _object: Optional[Node] = None,
    default: Optional[Any] = None,
    _any: Optional[bool] = True,
) -> Optional[Union[URIRef, Literal, BNode]]:
    # this is a helper method to cast some rdf types from graph.value() to be compatible with the
    # code that follows
    value = graph.value(subject=subject, predicate=predicate, object=_object, default=default, any=_any)
    if value != default and value is not None and not isinstance(value, (URIRef, Literal, BNode)):
        logger.append(
            f"Warning: Node {value} should be of type BNode, Literal or URIRef, but is {type(value).__name__}. "
            f"This might lead to a failure."
        )
    return value



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import pytest

from spdx_tools.spdx.model import Relationship, RelationshipType, SpdxNoAssertion


def test_correct_initialization():
    relationship = Relationship("id", RelationshipType.OTHER, SpdxNoAssertion(), "comment")
    assert relationship.spdx_element_id == "id"
    assert relationship.relationship_type == RelationshipType.OTHER
    assert relationship.related_spdx_element_id == SpdxNoAssertion()
    assert relationship.comment == "comment"


def test_wrong_type_in_spdx_element_id():
    with pytest.raises(TypeError):
        Relationship(SpdxNoAssertion(), RelationshipType.OTHER, "other_id")


def test_wrong_type_in_relationship_type():
    with pytest.raises(TypeError):
        Relationship("id", 42, "other_id")


def test_wrong_type_in_related_spdx_element_id():
    with pytest.raises(TypeError):
        Relationship("id", RelationshipType.OTHER, 42)


def test_wrong_type_in_comment():
    with pytest.raises(TypeError):
        Relationship("id", RelationshipType.OTHER, "other_id", 42)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from rdflib import RDF, RDFS, Graph, Literal, URIRef

from spdx_tools.spdx.datetime_conversions import datetime_to_iso_string
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE
from spdx_tools.spdx.writer.rdf.annotation_writer import add_annotation_to_graph
from tests.spdx.fixtures import annotation_fixture


def test_add_annotation_to_graph():
    graph = Graph()
    annotation = annotation_fixture()

    add_annotation_to_graph(annotation, graph, "docNamespace", {})

    assert (URIRef("docNamespace#SPDXRef-File"), SPDX_NAMESPACE.annotation, None) in graph
    assert (None, RDF.type, SPDX_NAMESPACE.Annotation) in graph
    assert (None, SPDX_NAMESPACE.annotationType, SPDX_NAMESPACE.annotationType_review) in graph
    assert (None, SPDX_NAMESPACE.annotationDate, Literal(datetime_to_iso_string(annotation.annotation_date))) in graph
    assert (None, SPDX_NAMESPACE.annotator, Literal(annotation.annotator.to_serialized_string())) in graph
    assert (None, RDFS.comment, Literal(annotation.annotation_comment)) in graph



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from spdx_tools.spdx.model.spdx_no_assertion import SpdxNoAssertion
from spdx_tools.spdx.model.spdx_none import SpdxNone
from spdx_tools.spdx.model.version import Version
from spdx_tools.spdx.model.actor import Actor, ActorType
from spdx_tools.spdx.model.annotation import Annotation, AnnotationType
from spdx_tools.spdx.model.checksum import Checksum, ChecksumAlgorithm
from spdx_tools.spdx.model.external_document_ref import ExternalDocumentRef
from spdx_tools.spdx.model.extracted_licensing_info import ExtractedLicensingInfo
from spdx_tools.spdx.model.file import File, FileType
from spdx_tools.spdx.model.package import (
    ExternalPackageRef,
    ExternalPackageRefCategory,
    Package,
    PackagePurpose,
    PackageVerificationCode,
)
from spdx_tools.spdx.model.relationship import Relationship, RelationshipType
from spdx_tools.spdx.model.snippet import Snippet
from spdx_tools.spdx.model.document import CreationInfo, Document



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from unittest import TestCase

import pytest

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.model import Actor, ActorType, Checksum, ChecksumAlgorithm, ExternalDocumentRef, Version
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.jsonlikedict.creation_info_parser import CreationInfoParser


def test_parse_creation_info():
    creation_info_parser = CreationInfoParser()
    doc_dict = {
        "spdxVersion": "2.3",
        "SPDXID": DOCUMENT_SPDX_ID,
        "name": "Example Document",
        "dataLicense": "CC0-1.0",
        "documentNamespace": "namespace",
        "externalDocumentRefs": [
            {
                "externalDocumentId": "DocumentRef-spdx-tool-1.2",
                "checksum": {"algorithm": "SHA1", "checksumValue": "d6a770ba38583ed4bb4525bd96e50461655d2759"},
                "spdxDocument": "http://spdx.org/spdxdocs/spdx-tools-v1.2-3F2504E0-4F89-41D3-9A0C-0305E82C3301",
            }
        ],
        "creationInfo": {
            "created": "2010-01-29T18:30:22Z",
            "creators": ["Tool: LicenseFind-1.0", "Organization: ExampleCodeInspect ()", "Person: Jane Doe ()"],
            "licenseListVersion": "3.7",
            "comment": "Some comment.",
        },
    }
    creation_info = creation_info_parser.parse_creation_info(doc_dict)

    assert creation_info.spdx_version == "2.3"
    assert creation_info.spdx_id == DOCUMENT_SPDX_ID
    assert creation_info.name == "Example Document"
    assert creation_info.document_namespace == "namespace"
    assert creation_info.created == datetime(2010, 1, 29, 18, 30, 22)
    TestCase().assertCountEqual(
        creation_info.creators,
        [
            Actor(ActorType.TOOL, "LicenseFind-1.0"),
            Actor(ActorType.ORGANIZATION, "ExampleCodeInspect"),
            Actor(ActorType.PERSON, "Jane Doe"),
        ],
    )
    assert creation_info.license_list_version == Version(3, 7)
    assert creation_info.external_document_refs == [
        ExternalDocumentRef(
            document_ref_id="DocumentRef-spdx-tool-1.2",
            checksum=Checksum(algorithm=ChecksumAlgorithm.SHA1, value="d6a770ba38583ed4bb4525bd96e50461655d2759"),
            document_uri="http://spdx.org/spdxdocs/spdx-tools-v1.2-3F2504E0-4F89-41D3-9A0C-0305E82C3301",
        )
    ]


@pytest.mark.parametrize(
    "incomplete_dict",
    [
        {"spdxVersion": "2.3", "SPDXID": DOCUMENT_SPDX_ID, "name": "Example Document"},
        {"creationInfo": {"created": "2019-02-01T11:30:40Z"}},
    ],
)
def test_parse_incomplete_document_info(incomplete_dict):
    creation_info_parser = CreationInfoParser()

    with pytest.raises(SPDXParsingError):
        creation_info_parser.parse_creation_info(incomplete_dict)


def test_parse_invalid_creation_info():
    creation_info_parser = CreationInfoParser()
    doc_dict = {
        "spdxVersion": "2.3",
        "SPDXID": DOCUMENT_SPDX_ID,
        "name": "Example Document",
        "creationInfo": {
            "created": "2010-01-29T18:30:22Z",
            "creators": ["Tool: LicenseFind-1.0", "Organization: ExampleCodeInspect ()", "Person: Jane Doe ()"],
        },
        "dataLicense": None,
    }

    with pytest.raises(SPDXParsingError):
        creation_info_parser.parse_creation_info(doc_dict)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest import TestCase

import pytest
from license_expression import Licensing

from spdx_tools.spdx.model import SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.jsonlikedict.snippet_parser import SnippetParser


@pytest.mark.parametrize(
    "copyright_text, expected_copyright_text",
    [
        ("Copyright 2008-2010 John Smith", "Copyright 2008-2010 John Smith"),
        ("NOASSERTION", SpdxNoAssertion()),
        ("NONE", SpdxNone()),
    ],
)
def test_parse_snippet(copyright_text, expected_copyright_text):
    snippet_parser = SnippetParser()

    snippet_dict = {
        "SPDXID": "SPDXRef-Snippet",
        "comment": "This snippet was identified as significant and highlighted in this Apache-2.0 file, when a "
        "commercial scanner identified it as being derived from file foo.c in package xyz which is licensed"
        " under GPL-2.0.",
        "copyrightText": copyright_text,
        "licenseComments": "The concluded license was taken from package xyz, from which the snippet was copied into "
        "the current file. The concluded license information was found in the COPYING.txt file in "
        "package xyz.",
        "licenseConcluded": "GPL-2.0-only",
        "licenseInfoInSnippets": ["GPL-2.0-only", "NOASSERTION"],
        "name": "from linux kernel",
        "ranges": [
            {
                "endPointer": {"offset": 420, "reference": "SPDXRef-DoapSource"},
                "startPointer": {"offset": 310, "reference": "SPDXRef-DoapSource"},
            },
            {
                "endPointer": {"lineNumber": 23, "reference": "SPDXRef-DoapSource"},
                "startPointer": {"lineNumber": 5, "reference": "SPDXRef-DoapSource"},
            },
        ],
        "snippetFromFile": "SPDXRef-DoapSource",
        "attributionTexts": ["Some example attibution text."],
    }
    snippet = snippet_parser.parse_snippet(snippet_dict)

    assert snippet.spdx_id == "SPDXRef-Snippet"
    assert snippet.name == "from linux kernel"
    assert (
        snippet.comment
        == "This snippet was identified as significant and highlighted in this Apache-2.0 file, when a commercial "
        "scanner identified it as being derived from file foo.c in package xyz which is licensed under GPL-2.0."
    )
    assert snippet.copyright_text == expected_copyright_text
    assert (
        snippet.license_comment
        == "The concluded license was taken from package xyz, from which the snippet was copied into the current file."
        " The concluded license information was found in the COPYING.txt file in package xyz."
    )
    assert snippet.byte_range == (310, 420)
    assert snippet.line_range == (5, 23)
    assert snippet.file_spdx_id == "SPDXRef-DoapSource"
    assert snippet.license_info_in_snippet == [Licensing().parse("GPL-2.0-only"), SpdxNoAssertion()]
    assert snippet.license_concluded == Licensing().parse("GPL-2.0-only")
    assert snippet.attribution_texts == ["Some example attibution text."]


def test_parse_incomplete_snippet():
    snippet_parser = SnippetParser()
    incomplete_snippet_dict = {"SPDXID": "SPDXRef-Snippet", "file_spdx_id": "SPDXRef-File"}

    with pytest.raises(SPDXParsingError):
        snippet_parser.parse_snippet(incomplete_snippet_dict)


def test_parse_snippet_with_invalid_snippet_range():
    snippet_parser = SnippetParser()
    snippet_with_invalid_ranges_list = {
        "SPDXID": "SPDXRef-Snippet",
        "file_spdx_id": "SPDXRef-File",
        "ranges": [
            {
                "endPointer": {"offset": 23, "reference": "SPDXRef-DoapSource"},
                "startPointer": {"offset": "310s", "reference": "SPDXRef-DoapSource"},
            }
        ],
    }

    with pytest.raises(SPDXParsingError):
        snippet_parser.parse_snippet(snippet_with_invalid_ranges_list)


def test_parse_invalid_snippet_range():
    snippet_parser = SnippetParser()

    ranges = [
        {
            "endPointer": {"lineNumber": 23, "reference": "SPDXRef-DoapSource"},
            "startPointer": {"offset": 310, "reference": "SPDXRef-DoapSource"},
        },
        {
            "endPointer": {"offset": 420, "reference": "SPDXRef-DoapSource"},
            "startPointer": {"lineNumber": 5, "reference": "SPDXRef-DoapSource"},
        },
    ]

    with pytest.raises(SPDXParsingError) as err:
        snippet_parser.parse_ranges(ranges)

    TestCase().assertCountEqual(
        err.value.get_messages(),
        [
            "Error while parsing snippet ranges: ['Type of startpointer is not the same as type of endpointer.', "
            "'Type of startpointer is not the same as type of endpointer.']"
        ],
    )



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import sys

import click
from beartype.typing import List

from spdx_tools.spdx3.bump_from_spdx2.spdx_document import bump_spdx_document
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx3.writer.console.payload_writer import write_payload as write_payload_to_console
from spdx_tools.spdx3.writer.json_ld.json_ld_writer import write_payload
from spdx_tools.spdx.model.document import Document
from spdx_tools.spdx.parser.parse_anything import parse_file
from spdx_tools.spdx.validation.document_validator import validate_full_spdx_document
from spdx_tools.spdx.validation.validation_message import ValidationMessage


@click.command()
@click.option(
    "--infile", "-i", prompt="input file path", help="The file containing the document to be validated or converted."
)
@click.option(
    "--outfile",
    "-o",
    help="The file to write the converted document to (write a dash for output to stdout or omit for no conversion)."
    "For now only a prototype serialization to json-ld is available. The provided file will therefore be extended "
    "with `.jsonld`.",
)
@click.option(
    "--version",
    help='The SPDX version to be used during parsing and validation (format "SPDX-2.3").',
    default="SPDX-2.3",
)
@click.option("--novalidation", is_flag=True, help="Don't validate the provided document.")
def main(infile: str, outfile: str, version: str, novalidation: bool):
    """
    CLI-tool to parse and validate a SPDX 2.x document and migrate it into the prototype of SPDX 3.0. As there is no
    definition for a serialization yet output can only be written to stdout.
    To use, run: 'pyspdxtools3 --infile <input file name> -o -'
    """
    try:
        document: Document = parse_file(infile)

        if not novalidation:
            validation_messages: List[ValidationMessage] = validate_full_spdx_document(document, version)
            if validation_messages:
                print("The document is invalid. The following issues have been found:", file=sys.stderr)
                for message in validation_messages:
                    print(message.validation_message, file=sys.stderr)
                sys.exit(1)
            else:
                print("The document is valid.", file=sys.stderr)
        if outfile:
            payload: Payload = bump_spdx_document(document)
            if outfile == "-":
                write_payload_to_console(payload, sys.stdout)
            else:
                write_payload(payload, outfile)

    except NotImplementedError as err:
        print(err.args[0])
        sys.exit(1)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from rdflib import RDF, RDFS, Graph, Literal, URIRef

from spdx_tools.spdx.rdfschema.namespace import LICENSE_NAMESPACE, SPDX_NAMESPACE
from spdx_tools.spdx.writer.rdf.file_writer import add_file_to_graph
from tests.spdx.fixtures import file_fixture


def test_add_file_to_graph():
    graph = Graph()
    file = file_fixture()

    add_file_to_graph(file, graph, "docNamespace", {})

    assert (URIRef("docNamespace#SPDXRef-File"), RDF.type, SPDX_NAMESPACE.File) in graph
    assert (None, SPDX_NAMESPACE.fileName, Literal(file.name)) in graph
    assert (None, SPDX_NAMESPACE.fileType, SPDX_NAMESPACE.fileType_text) in graph
    assert (None, SPDX_NAMESPACE.licenseComments, Literal(file.license_comment)) in graph
    assert (None, SPDX_NAMESPACE.licenseConcluded, None) in graph
    assert (None, SPDX_NAMESPACE.licenseInfoInFile, LICENSE_NAMESPACE.MIT) in graph
    assert (None, SPDX_NAMESPACE.licenseInfoInFile, LICENSE_NAMESPACE["GPL-2.0-only"]) in graph
    assert (None, SPDX_NAMESPACE.licenseInfoInFile, SPDX_NAMESPACE.noassertion) in graph
    assert (None, SPDX_NAMESPACE.copyrightText, Literal(file.copyright_text)) in graph
    assert (None, RDFS.comment, Literal(file.comment)) in graph
    assert (None, SPDX_NAMESPACE.noticeText, Literal(file.notice)) in graph
    assert (None, SPDX_NAMESPACE.fileContributor, Literal(file.contributors[0])) in graph
    assert (None, SPDX_NAMESPACE.checksum, None) in graph
    assert (None, SPDX_NAMESPACE.attributionText, Literal(file.attribution_texts[0])) in graph



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import List, Union

from spdx_tools.spdx.model import ExternalDocumentRef, File, Package, Snippet


def get_full_element_spdx_id(
    element: Union[Package, File, Snippet],
    document_namespace: str,
    external_document_refs: List[ExternalDocumentRef],
) -> str:
    """
    Returns the spdx_id of the element prefixed with the correct document namespace and,
    if the element is from an external document, sets the correct entry in the imports property.
    """
    if ":" not in element.spdx_id:
        return f"{document_namespace}#{element.spdx_id}"

    external_id, local_id = element.spdx_id.split(":")
    external_uri = None
    for entry in external_document_refs:
        if entry.document_ref_id == external_id:
            external_uri = entry.document_uri
            break

    if not external_uri:
        raise ValueError(f"external id {external_id} not found in external document references")

    return external_uri + "#" + local_id



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime
from unittest import mock

import pytest

from spdx_tools.spdx.model import Annotation, AnnotationType


@mock.patch("spdx_tools.spdx.model.Actor", autospec=True)
def test_correct_initialization(actor):
    annotation = Annotation("id", AnnotationType.OTHER, actor, datetime(2022, 1, 1), "comment")
    assert annotation.spdx_id == "id"
    assert annotation.annotation_type == AnnotationType.OTHER
    assert annotation.annotator == actor
    assert annotation.annotation_date == datetime(2022, 1, 1)
    assert annotation.annotation_comment == "comment"


@mock.patch("spdx_tools.spdx.model.Actor", autospec=True)
def test_wrong_type_in_spdx_id(actor):
    with pytest.raises(TypeError):
        Annotation(42, AnnotationType.OTHER, actor, datetime(2022, 1, 1), "comment")


@mock.patch("spdx_tools.spdx.model.Actor", autospec=True)
def test_wrong_type_in_annotation_type(actor):
    with pytest.raises(TypeError):
        Annotation("id", 42, actor, datetime(2022, 1, 1), "comment")


@mock.patch("spdx_tools.spdx.model.Actor", autospec=True)
def test_wrong_type_in_annotator(actor):
    with pytest.raises(TypeError):
        Annotation("id", AnnotationType.OTHER, 42, datetime(2022, 1, 1), "comment")


@mock.patch("spdx_tools.spdx.model.Actor", autospec=True)
def test_wrong_type_in_annotation_date(actor):
    with pytest.raises(TypeError):
        Annotation("id", AnnotationType.OTHER, actor, 42, "comment")


@mock.patch("spdx_tools.spdx.model.Actor", autospec=True)
def test_wrong_type_in_annotation_comment(actor):
    with pytest.raises(TypeError):
        Annotation("id", AnnotationType.OTHER, actor, datetime(2022, 1, 1), 42)



# SPDX-License-Identifier: Apache-2.0
#  Copyright (c) 2022 spdx contributors
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#    http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from beartype.typing import TextIO

from spdx_tools.spdx.model import File
from spdx_tools.spdx.writer.tagvalue.checksum_writer import write_checksum_to_tag_value
from spdx_tools.spdx.writer.tagvalue.tagvalue_writer_helper_functions import write_text_value, write_value


def write_file(file: File, text_output: TextIO):
    text_output.write("## File Information\n")

    write_value("FileName", file.name, text_output)
    write_value("SPDXID", file.spdx_id, text_output)

    for file_type in file.file_types:
        write_value("FileType", file_type.name, text_output)

    for file_checksum in file.checksums:
        write_value("FileChecksum", write_checksum_to_tag_value(file_checksum), text_output)

    write_value("LicenseConcluded", file.license_concluded, text_output)
    for license_info in file.license_info_in_file:
        write_value("LicenseInfoInFile", license_info, text_output)
    write_text_value("LicenseComments", file.license_comment, text_output)
    write_text_value("FileCopyrightText", file.copyright_text, text_output)

    write_text_value("FileComment", file.comment, text_output)
    write_text_value("FileNotice", file.notice, text_output)

    for contributor in sorted(file.contributors):
        write_value("FileContributor", contributor, text_output)

    for attribution_text in file.attribution_texts:
        write_text_value("FileAttributionText", attribution_text, text_output)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from spdx_tools.spdx3.bump_from_spdx2.checksum import bump_checksum
from spdx_tools.spdx3.bump_from_spdx2.creation_info import bump_creation_info
from spdx_tools.spdx3.bump_from_spdx2.external_document_ref import bump_external_document_ref
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx.model import ExternalDocumentRef
from tests.spdx.fixtures import checksum_fixture, creation_info_fixture, external_document_ref_fixture


def test_bump_external_document_ref():
    checksum = checksum_fixture()
    external_document_ref = ExternalDocumentRef("DocumentRef-external", "https://external.uri", checksum)
    namespace_map, imports = bump_external_document_ref(external_document_ref)

    assert namespace_map.prefix == "DocumentRef-external"
    assert namespace_map.namespace == "https://external.uri#"

    assert imports.external_id == "DocumentRef-external:SPDXRef-DOCUMENT"
    assert imports.verified_using == [bump_checksum(checksum)]


def test_bump_multiple_external_document_refs():
    payload = Payload()
    creation_info = creation_info_fixture(
        external_document_refs=[
            external_document_ref_fixture("DocumentRef-external1", "https://external.uri1"),
            external_document_ref_fixture("DocumentRef-external2", "https://external.uri2"),
        ]
    )
    spdx_document = bump_creation_info(creation_info, payload)

    assert len(spdx_document.imports) == 2
    assert len(spdx_document.namespaces) == 2



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import List, Optional, Tuple

from spdx_tools.spdx3.bump_from_spdx2.message import print_missing_conversion
from spdx_tools.spdx3.model import ExternalMap
from spdx_tools.spdx3.model.positive_integer_range import PositiveIntegerRange
from spdx_tools.spdx3.model.software import Snippet
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx.model import ExternalDocumentRef, SpdxNoAssertion
from spdx_tools.spdx.model.snippet import Snippet as Spdx2_Snippet
from spdx_tools.spdx.spdx_element_utils import get_full_element_spdx_id


def bump_integer_range(spdx2_range: Optional[Tuple[int, int]]) -> PositiveIntegerRange:
    return PositiveIntegerRange(spdx2_range[0], spdx2_range[1]) if spdx2_range else None


def bump_snippet(
    spdx2_snippet: Spdx2_Snippet,
    payload: Payload,
    document_namespace: str,
    external_document_refs: List[ExternalDocumentRef],
    imports: List[ExternalMap],
):
    spdx_id = get_full_element_spdx_id(spdx2_snippet, document_namespace, external_document_refs)
    if ":" in spdx2_snippet.spdx_id:
        imports.append(
            ExternalMap(
                external_id=spdx2_snippet.spdx_id,
                defining_document=f"{spdx2_snippet.spdx_id.split(':')[0]}:SPDXRef-DOCUMENT",
            )
        )

    print_missing_conversion("snippet.file_spdx_id", 0, "https://github.com/spdx/spdx-3-model/issues/130")
    copyright_text = None
    if isinstance(spdx2_snippet.copyright_text, str):
        copyright_text = spdx2_snippet.copyright_text
    elif isinstance(spdx2_snippet.copyright_text, SpdxNoAssertion):
        print_missing_conversion("package2.copyright_text", 0)
    print_missing_conversion(
        "snippet.license_info_in_snippet, snippet.license_comment,",
        0,
        "missing definitions for license profile",
    )

    payload.add_element(
        Snippet(
            spdx_id=spdx_id,
            name=spdx2_snippet.name,
            comment=spdx2_snippet.comment,
            byte_range=bump_integer_range(spdx2_snippet.byte_range),
            line_range=bump_integer_range(spdx2_snippet.line_range),
            copyright_text=copyright_text,
            attribution_text=", ".join(spdx2_snippet.attribution_texts),
        )
    )



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import yaml
from beartype.typing import Dict

from spdx_tools.spdx.model import Document
from spdx_tools.spdx.parser.jsonlikedict.json_like_dict_parser import JsonLikeDictParser


def parse_from_file(file_name: str) -> Document:
    with open(file_name) as file:
        input_doc_as_dict: Dict = yaml.safe_load(file)

    return JsonLikeDictParser().parse(input_doc_as_dict)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import auto

from spdx_tools.spdx.jsonschema.json_property import JsonProperty


class PackageProperty(JsonProperty):
    SPDX_ID = auto()
    ANNOTATIONS = auto()
    ATTRIBUTION_TEXTS = auto()
    BUILT_DATE = auto()
    CHECKSUMS = auto()
    COMMENT = auto()
    COPYRIGHT_TEXT = auto()
    DESCRIPTION = auto()
    DOWNLOAD_LOCATION = auto()
    EXTERNAL_REFS = auto()
    FILES_ANALYZED = auto()
    HOMEPAGE = auto()
    LICENSE_COMMENTS = auto()
    LICENSE_CONCLUDED = auto()
    LICENSE_DECLARED = auto()
    LICENSE_INFO_FROM_FILES = auto()
    NAME = auto()
    ORIGINATOR = auto()
    PACKAGE_FILE_NAME = auto()
    PACKAGE_VERIFICATION_CODE = auto()
    PRIMARY_PACKAGE_PURPOSE = auto()
    RELEASE_DATE = auto()
    SOURCE_INFO = auto()
    SUMMARY = auto()
    SUPPLIER = auto()
    VALID_UNTIL_DATE = auto()
    VERSION_INFO = auto()



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx3.model.licensing import ConjunctiveLicenseSet, ListedLicense
from tests.spdx3.fixtures import fixture_factory


def test_valid_initialization():
    lic = fixture_factory(ListedLicense)
    conjunctive_license_set = ConjunctiveLicenseSet([lic, lic])

    assert conjunctive_license_set.member == [lic, lic]


def test_invalid_initialization():
    lic = fixture_factory(ListedLicense)
    with pytest.raises(TypeError) as err:
        ConjunctiveLicenseSet(lic)

    assert len(err.value.args[0]) == 1
    assert err.value.args[0][0].startswith("SetterError ConjunctiveLicenseSet:")



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx.jsonschema.checksum_converter import ChecksumConverter
from spdx_tools.spdx.jsonschema.checksum_properties import ChecksumProperty
from spdx_tools.spdx.model import Checksum, ChecksumAlgorithm


@pytest.fixture
def converter() -> ChecksumConverter:
    return ChecksumConverter()


@pytest.mark.parametrize(
    "checksum_property,expected",
    [(ChecksumProperty.ALGORITHM, "algorithm"), (ChecksumProperty.CHECKSUM_VALUE, "checksumValue")],
)
def test_json_property_names(converter: ChecksumConverter, checksum_property: ChecksumProperty, expected: str):
    assert converter.json_property_name(checksum_property) == expected


def test_successful_conversion(converter: ChecksumConverter):
    checksum = Checksum(ChecksumAlgorithm.SHA1, "123")

    converted_dict = converter.convert(checksum)

    assert converted_dict == {
        converter.json_property_name(ChecksumProperty.ALGORITHM): "SHA1",
        converter.json_property_name(ChecksumProperty.CHECKSUM_VALUE): "123",
    }


def test_json_type(converter: ChecksumConverter):
    assert converter.get_json_type() == ChecksumProperty


def test_data_model_type(converter: ChecksumConverter):
    assert converter.get_data_model_type() == Checksum



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import pytest

from spdx_tools.spdx.model import Snippet, SpdxNoAssertion, SpdxNone


def test_correct_initialization():
    snippet = Snippet(
        "id",
        "file_id",
        (200, 400),
        (20, 40),
        SpdxNone(),
        [SpdxNoAssertion()],
        "comment on license",
        "copyright",
        "comment",
        "name",
        ["attribution"],
    )
    assert snippet.spdx_id == "id"
    assert snippet.file_spdx_id == "file_id"
    assert snippet.byte_range == (200, 400)
    assert snippet.line_range == (20, 40)
    assert snippet.license_concluded == SpdxNone()
    assert snippet.license_info_in_snippet == [SpdxNoAssertion()]
    assert snippet.license_comment == "comment on license"
    assert snippet.copyright_text == "copyright"
    assert snippet.comment == "comment"
    assert snippet.name == "name"
    assert snippet.attribution_texts == ["attribution"]


def test_correct_initialization_with_default_values():
    snippet = Snippet("id", "file_id", (200, 400))
    assert snippet.spdx_id == "id"
    assert snippet.file_spdx_id == "file_id"
    assert snippet.byte_range == (200, 400)
    assert snippet.line_range is None
    assert snippet.license_concluded is None
    assert snippet.license_info_in_snippet == []
    assert snippet.license_comment is None
    assert snippet.copyright_text is None
    assert snippet.comment is None
    assert snippet.name is None
    assert snippet.attribution_texts == []


def test_wrong_type_in_spdx_id():
    with pytest.raises(TypeError):
        Snippet(42, "file_id", (200, 400))


def test_wrong_type_in_file_spdx_id():
    with pytest.raises(TypeError):
        Snippet("id", 42, (200, 400))


def test_wrong_type_in_byte_range():
    with pytest.raises(TypeError):
        Snippet("id", "file_id", (200, 300, 400))


def test_wrong_type_in_line_range():
    with pytest.raises(TypeError):
        Snippet("id", "file_id", (200, 400), line_range=(20, "40"))


def test_wrong_type_in_license_concluded():
    with pytest.raises(TypeError):
        Snippet("id", "file_id", (200, 400), license_concluded="NONE")


def test_wrong_type_in_license_info_in_snippet():
    with pytest.raises(TypeError):
        Snippet("id", "file_id", (200, 400), license_info_in_snippet=SpdxNoAssertion())


def test_wrong_type_in_license_comment():
    with pytest.raises(TypeError):
        Snippet("id", "file_id", (200, 400), license_comment=[])


def test_wrong_type_in_copyright_text():
    with pytest.raises(TypeError):
        Snippet("id", "file_id", (200, 400), copyright_text=["copyright"])


def test_wrong_type_in_comment():
    with pytest.raises(TypeError):
        Snippet("id", "file_id", (200, 400), comment=["comment"])


def test_wrong_type_in_name():
    with pytest.raises(TypeError):
        Snippet("id", "file_id", (200, 400), name=42)


def test_wrong_type_in_attribution_texts():
    with pytest.raises(TypeError):
        Snippet("id", "file_id", (200, 400), attribution_texts="attribution")



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import field

from beartype.typing import List, Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx3.model import IntegrityMethod


@dataclass_with_properties
class ExternalMap:
    external_id: str  # anyURI
    verified_using: List[IntegrityMethod] = field(default_factory=list)
    location_hint: Optional[str] = None  # anyURI
    defining_document: Optional[str] = None

    def __init__(
        self,
        external_id: str,
        verified_using: List[IntegrityMethod] = None,
        location_hint: Optional[str] = None,
        defining_document: Optional[str] = None,
    ):
        verified_using = [] if verified_using is None else verified_using
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import re

from beartype.typing import Dict, List

from spdx_tools.spdx.model import Checksum, ChecksumAlgorithm
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage

# in hexadecimal digits
algorithm_length: Dict = {
    ChecksumAlgorithm.SHA1: "40",
    ChecksumAlgorithm.SHA224: "56",
    ChecksumAlgorithm.SHA256: "64",
    ChecksumAlgorithm.SHA384: "96",
    ChecksumAlgorithm.SHA512: "128",
    ChecksumAlgorithm.SHA3_256: "64",
    ChecksumAlgorithm.SHA3_384: "96",
    ChecksumAlgorithm.SHA3_512: "128",
    ChecksumAlgorithm.BLAKE2B_256: "64",
    ChecksumAlgorithm.BLAKE2B_384: "96",
    ChecksumAlgorithm.BLAKE2B_512: "128",
    ChecksumAlgorithm.BLAKE3: "256,",  # at least 256 bits
    ChecksumAlgorithm.MD2: "32",
    ChecksumAlgorithm.MD4: "32",
    ChecksumAlgorithm.MD5: "32",
    ChecksumAlgorithm.MD6: "0,512",  # between 0 and 512 bits
    ChecksumAlgorithm.ADLER32: "8",
}


def validate_checksums(checksums: List[Checksum], parent_id: str, spdx_version: str) -> List[ValidationMessage]:
    validation_messages = []
    for checksum in checksums:
        validation_messages.extend(validate_checksum(checksum, parent_id, spdx_version))

    return validation_messages


def validate_checksum(checksum: Checksum, parent_id: str, spdx_version: str) -> List[ValidationMessage]:
    validation_messages = []
    algorithm = checksum.algorithm
    context = ValidationContext(parent_id=parent_id, element_type=SpdxElementType.CHECKSUM, full_element=checksum)

    if spdx_version == "SPDX-2.2" and algorithm in [
        ChecksumAlgorithm.SHA3_512,
        ChecksumAlgorithm.SHA3_384,
        ChecksumAlgorithm.SHA3_256,
        ChecksumAlgorithm.BLAKE3,
        ChecksumAlgorithm.BLAKE2B_512,
        ChecksumAlgorithm.BLAKE2B_384,
        ChecksumAlgorithm.BLAKE2B_256,
        ChecksumAlgorithm.ADLER32,
    ]:
        return [ValidationMessage(f"{checksum.algorithm.name} is not supported in SPDX-2.2", context)]

    if not re.match("^[0-9a-f]{" + algorithm_length[algorithm] + "}$", checksum.value):
        if algorithm == ChecksumAlgorithm.BLAKE3:
            length = "at least 256"
        elif algorithm == ChecksumAlgorithm.MD6:
            length = "between 0 and 512"
        else:
            length = algorithm_length[algorithm]
        validation_messages.append(
            ValidationMessage(
                f"value of {algorithm} must consist of {length} lowercase hexadecimal digits, but is: "
                f"{checksum.value} (length: {len(checksum.value)} digits)",
                context,
            )
        )

    return validation_messages



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import Enum, auto

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values


class ChecksumAlgorithm(Enum):
    SHA1 = auto()
    SHA224 = auto()
    SHA256 = auto()
    SHA384 = auto()
    SHA512 = auto()
    SHA3_256 = auto()
    SHA3_384 = auto()
    SHA3_512 = auto()
    BLAKE2B_256 = auto()
    BLAKE2B_384 = auto()
    BLAKE2B_512 = auto()
    BLAKE3 = auto()
    MD2 = auto()
    MD4 = auto()
    MD5 = auto()
    MD6 = auto()
    ADLER32 = auto()


@dataclass_with_properties
class Checksum:
    algorithm: ChecksumAlgorithm
    value: str

    def __init__(self, algorithm: ChecksumAlgorithm, value: str):
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from spdx_tools.spdx3.bump_from_spdx2.snippet import bump_snippet
from spdx_tools.spdx3.model.software import Snippet
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx.model.snippet import Snippet as Spdx2_Snippet
from tests.spdx.fixtures import snippet_fixture


def test_bump_snippet():
    payload = Payload()
    document_namespace = "https://doc.namespace"
    spdx2_snippet: Spdx2_Snippet = snippet_fixture()
    expected_new_snippet_id = f"{document_namespace}#{spdx2_snippet.spdx_id}"

    bump_snippet(spdx2_snippet, payload, document_namespace, [], [])
    snippet = payload.get_element(expected_new_snippet_id)

    assert isinstance(snippet, Snippet)
    assert snippet.spdx_id == expected_new_snippet_id
    assert snippet.copyright_text == spdx2_snippet.copyright_text
    assert snippet.attribution_text == spdx2_snippet.attribution_texts[0]



#  SPDX-FileCopyrightText: 2023 spdx contributors
#
#  SPDX-License-Identifier: Apache-2.0
from os import path

from spdx_tools.spdx.model import Document
from spdx_tools.spdx.writer.write_anything import write_file
from spdx_tools.spdx.parser.parse_anything import parse_file

# This example demonstrates how to load an existing SPDX2 file and convert it to a different SPDX2 format

# Provide a path to the input file in the originating format
input_path = path.join(path.dirname(__file__), "..", "tests", "spdx", "data", "SPDXLite.spdx")
# Parse the original input file (format is deduced automatically from the file extension)
document: Document = parse_file(input_path)
# Write to a different file format (e.g. XML, format is deduced automatically from the file extension)
write_file(document, "converted_format.xml")



# SPDX-FileCopyrightText: 2023 SPDX Contributors
#
# SPDX-License-Identifier: Apache-2.0
import datetime
from unittest.mock import MagicMock, call, mock_open, patch

import pytest

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.model import CreationInfo
from spdx_tools.spdx.writer.tagvalue.creation_info_writer import write_creation_info
from tests.spdx.fixtures import actor_fixture, creation_info_fixture


@pytest.mark.parametrize(
    "creation_info, expected_calls",
    [
        (
            creation_info_fixture(),
            [
                call("SPDXVersion: SPDX-2.3\n"),
                call("DataLicense: CC0-1.0\n"),
                call(f"SPDXID: {DOCUMENT_SPDX_ID}\n"),
                call("DocumentName: documentName\n"),
                call("DocumentNamespace: https://some.namespace\n"),
                call("DocumentComment: documentComment\n"),
                call("\n## External Document References\n"),
                call(
                    "ExternalDocumentRef: DocumentRef-external https://namespace.com "
                    "SHA1: 71c4025dd9897b364f3ebbb42c484ff43d00791c\n"
                ),
                call("\n"),
                call("## Creation Information\n"),
                call("LicenseListVersion: 3.19\n"),
                call("Creator: Person: creatorName (some@mail.com)\n"),
                call("Created: 2022-12-01T00:00:00Z\n"),
                call("CreatorComment: creatorComment\n"),
            ],
        ),
        (
            CreationInfo(
                spdx_version="SPDX-2.3",
                spdx_id=DOCUMENT_SPDX_ID,
                creators=[actor_fixture()],
                name="Test document",
                document_namespace="https://namespace.com",
                created=datetime.datetime(2022, 3, 10),
            ),
            [
                call("SPDXVersion: SPDX-2.3\n"),
                call("DataLicense: CC0-1.0\n"),
                call(f"SPDXID: {DOCUMENT_SPDX_ID}\n"),
                call("DocumentName: Test document\n"),
                call("DocumentNamespace: https://namespace.com\n"),
                call("\n"),
                call("## Creation Information\n"),
                call("Creator: Person: actorName (some@mail.com)\n"),
                call("Created: 2022-03-10T00:00:00Z\n"),
            ],
        ),
    ],
)
def test_creation_info_writer(creation_info, expected_calls):
    mock: MagicMock = mock_open()
    with patch(f"{__name__}.open", mock, create=True):
        with open("foo", "w") as file:
            write_creation_info(creation_info, file)

    mock.assert_called_once_with("foo", "w")
    handle = mock()
    handle.write.assert_has_calls(expected_calls)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime

from beartype.typing import Dict, List, Optional

from spdx_tools.spdx.datetime_conversions import datetime_from_str
from spdx_tools.spdx.model import Actor, Annotation, AnnotationType
from spdx_tools.spdx.parser.actor_parser import ActorParser
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.jsonlikedict.dict_parsing_functions import (
    append_parsed_field_or_log_error,
    parse_field_or_log_error,
)
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import (
    construct_or_raise_parsing_error,
    raise_parsing_error_if_logger_has_messages,
)


class AnnotationParser:
    logger: Logger
    actor_parser: ActorParser

    def __init__(self):
        self.logger = Logger()
        self.actor_parser = ActorParser()

    def parse_all_annotations(self, input_doc_dict: Dict) -> List[Annotation]:
        annotations = []
        self.parse_annotations_from_object(annotations, [input_doc_dict])
        reviews: List[Dict] = input_doc_dict.get("revieweds", [])
        for review in reviews:
            annotations = append_parsed_field_or_log_error(
                self.logger, annotations, review, lambda x: self.parse_review(x, spdx_id=input_doc_dict.get("SPDXID"))
            )
        packages: List[Dict] = input_doc_dict.get("packages", [])
        self.parse_annotations_from_object(annotations, packages)
        files: List[Dict] = input_doc_dict.get("files", [])
        self.parse_annotations_from_object(annotations, files)
        snippets: List[Dict] = input_doc_dict.get("snippets", [])
        self.parse_annotations_from_object(annotations, snippets)

        raise_parsing_error_if_logger_has_messages(self.logger, "annotations")
        return annotations

    def parse_annotations_from_object(self, annotations: List[Annotation], element_list: List[Dict]):
        for element in element_list:
            element_spdx_id: Optional[str] = element.get("SPDXID")
            element_annotations: List[Dict] = element.get("annotations", [])
            annotations.extend(
                parse_field_or_log_error(
                    self.logger,
                    element_annotations,
                    lambda y: self.parse_annotation(y, spdx_id=element_spdx_id),
                    [],
                    True,
                )
            )

    def parse_annotation(self, annotation_dict: Dict, spdx_id: Optional[str] = None) -> Annotation:
        logger = Logger()
        spdx_id: Optional[str] = annotation_dict.get("SPDXID") or spdx_id

        annotation_type: Optional[AnnotationType] = parse_field_or_log_error(
            logger, annotation_dict.get("annotationType"), self.parse_annotation_type
        )

        annotator: Optional[Actor] = parse_field_or_log_error(
            logger, annotation_dict.get("annotator"), self.actor_parser.parse_actor
        )

        annotation_date: Optional[datetime] = parse_field_or_log_error(
            logger, annotation_dict.get("annotationDate"), datetime_from_str
        )

        annotation_comment: Optional[str] = annotation_dict.get("comment")
        raise_parsing_error_if_logger_has_messages(logger, "Annotation")
        annotation_dict = construct_or_raise_parsing_error(
            Annotation,
            dict(
                spdx_id=spdx_id,
                annotation_type=annotation_type,
                annotator=annotator,
                annotation_date=annotation_date,
                annotation_comment=annotation_comment,
            ),
        )

        return annotation_dict

    @staticmethod
    def parse_annotation_type(annotation_type: str) -> AnnotationType:
        try:
            return AnnotationType[annotation_type]
        except KeyError:
            raise SPDXParsingError([f"Invalid AnnotationType: {annotation_type}"])

    def parse_review(self, review_dict: Dict, spdx_id: str) -> Annotation:
        logger = Logger()
        annotator: Optional[Actor] = parse_field_or_log_error(
            logger, review_dict.get("reviewer"), self.actor_parser.parse_actor
        )

        annotation_date: Optional[datetime] = parse_field_or_log_error(
            logger, review_dict.get("reviewDate"), datetime_from_str
        )

        annotation_type = AnnotationType.REVIEW
        comment: Optional[str] = review_dict.get("comment")
        raise_parsing_error_if_logger_has_messages(logger, "Annotation from revieweds")

        annotation = construct_or_raise_parsing_error(
            Annotation,
            dict(
                spdx_id=spdx_id,
                annotation_type=annotation_type,
                annotator=annotator,
                annotation_date=annotation_date,
                annotation_comment=comment,
            ),
        )
        return annotation



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import Enum, auto

from beartype.typing import Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values


class ActorType(Enum):
    PERSON = auto()
    ORGANIZATION = auto()
    TOOL = auto()


@dataclass_with_properties
class Actor:
    actor_type: ActorType
    name: str
    email: Optional[str] = None

    def __init__(self, actor_type: ActorType, name: str, email: Optional[str] = None):
        check_types_and_set_values(self, locals())

    def to_serialized_string(self) -> str:
        """
        All serialization formats use the same representation of an actor, so this method is included in the data model
        """
        optional_email = f" ({self.email})" if self.email else ""
        return "".join([f"{self.actor_type.name.title()}:", f" {self.name}", optional_email])

    def __str__(self):
        return self.to_serialized_string()



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import logging
import sys

from beartype.typing import Dict, List, Optional, Tuple, Union

from spdx_tools.spdx3.bump_from_spdx2.message import print_missing_conversion
from spdx_tools.spdx3.model import LifecycleScopeType, Relationship, RelationshipCompleteness, RelationshipType
from spdx_tools.spdx3.model.software import (
    DependencyConditionalityType,
    SoftwareDependencyLinkType,
    SoftwareDependencyRelationship,
)
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx.model.relationship import Relationship as Spdx2_Relationship
from spdx_tools.spdx.model.relationship import RelationshipType as Spdx2_RelationshipType
from spdx_tools.spdx.model.spdx_no_assertion import SpdxNoAssertion
from spdx_tools.spdx.model.spdx_none import SpdxNone

# bump relationship type, map each relationship type to the corresponding class in 3.0,
# the relationship type, other arguments and if swapped
relationship_mapping: Dict[
    Spdx2_RelationshipType,
    Tuple[
        Union[Relationship, SoftwareDependencyRelationship],
        RelationshipType,
        Dict[str, Union[bool, LifecycleScopeType, SoftwareDependencyLinkType, DependencyConditionalityType]],
    ],
] = {
    Spdx2_RelationshipType.AMENDS: (Relationship, RelationshipType.AMENDS, {}),
    Spdx2_RelationshipType.ANCESTOR_OF: (Relationship, RelationshipType.ANCESTOR, {}),
    Spdx2_RelationshipType.BUILD_DEPENDENCY_OF: (
        SoftwareDependencyRelationship,
        RelationshipType.DEPENDS_ON,
        {
            "scope": LifecycleScopeType.BUILD,
            "linkage": SoftwareDependencyLinkType.TOOL,
        },
    ),
    Spdx2_RelationshipType.BUILD_TOOL_OF: (
        SoftwareDependencyRelationship,
        RelationshipType.DEPENDS_ON,
        {"scope": LifecycleScopeType.BUILD, "linkage": SoftwareDependencyLinkType.TOOL},
    ),
    Spdx2_RelationshipType.CONTAINED_BY: (Relationship, RelationshipType.CONTAINS, {"swap": True}),
    Spdx2_RelationshipType.CONTAINS: (
        Relationship,
        RelationshipType.CONTAINS,
        {},
    ),  # might be deleted in favor of depends on
    Spdx2_RelationshipType.COPY_OF: (Relationship, RelationshipType.COPY, {}),
    Spdx2_RelationshipType.DATA_FILE_OF: (None, None, {}),  # not defined, probably input/ output
    Spdx2_RelationshipType.DEPENDENCY_MANIFEST_OF: (
        SoftwareDependencyRelationship,
        RelationshipType.DEPENDS_ON,
        {},
    ),  # "expect purpose has been set to manifest"
    Spdx2_RelationshipType.DEPENDENCY_OF: (
        SoftwareDependencyRelationship,
        RelationshipType.DEPENDS_ON,
        {"swap": True},
    ),
    Spdx2_RelationshipType.DEPENDS_ON: (SoftwareDependencyRelationship, RelationshipType.DEPENDS_ON, {}),
    Spdx2_RelationshipType.DESCENDANT_OF: (Relationship, RelationshipType.ANCESTOR, {"swap": True}),
    Spdx2_RelationshipType.DESCRIBED_BY: (Relationship, RelationshipType.DESCRIBES, {"swap": True}),
    Spdx2_RelationshipType.DESCRIBES: (
        Relationship,
        RelationshipType.DESCRIBES,
        {},
    ),  # might be deleted in favor of root
    # property
    Spdx2_RelationshipType.DEV_DEPENDENCY_OF: (
        SoftwareDependencyRelationship,
        RelationshipType.DEPENDS_ON,
        {"scope": LifecycleScopeType.DEVELOPMENT},
    ),
    Spdx2_RelationshipType.DEV_TOOL_OF: (
        SoftwareDependencyRelationship,
        RelationshipType.DEPENDS_ON,
        {"scope": LifecycleScopeType.DEVELOPMENT, "linkage": SoftwareDependencyLinkType.TOOL},
    ),
    Spdx2_RelationshipType.DISTRIBUTION_ARTIFACT: (None, None, {}),  # not defined yet, purpose?
    Spdx2_RelationshipType.DOCUMENTATION_OF: (Relationship, RelationshipType.DOCUMENTATION, {}),
    Spdx2_RelationshipType.DYNAMIC_LINK: (
        SoftwareDependencyRelationship,
        RelationshipType.DEPENDS_ON,
        {"linkage": SoftwareDependencyLinkType.DYNAMIC},
    ),
    Spdx2_RelationshipType.EXAMPLE_OF: (Relationship, RelationshipType.EXAMPLE, {}),
    Spdx2_RelationshipType.EXPANDED_FROM_ARCHIVE: (Relationship, RelationshipType.EXPANDED_FROM_ARCHIVE, {}),
    Spdx2_RelationshipType.FILE_ADDED: (Relationship, RelationshipType.FILE_ADDED, {}),
    Spdx2_RelationshipType.FILE_DELETED: (Relationship, RelationshipType.FILE_DELETED, {}),
    Spdx2_RelationshipType.FILE_MODIFIED: (Relationship, RelationshipType.FILE_MODIFIED, {}),
    Spdx2_RelationshipType.GENERATED_FROM: (Relationship, RelationshipType.GENERATES, {"swap": True}),
    Spdx2_RelationshipType.GENERATES: (Relationship, RelationshipType.GENERATES, {}),
    Spdx2_RelationshipType.HAS_PREREQUISITE: (
        SoftwareDependencyRelationship,
        RelationshipType.DEPENDS_ON,
        {"conditionality": DependencyConditionalityType.PREREQUISITE},
    ),
    Spdx2_RelationshipType.METAFILE_OF: (Relationship, RelationshipType.METAFILE, {}),
    Spdx2_RelationshipType.OPTIONAL_COMPONENT_OF: (None, None, {}),  # converted to depends on and purpose? not clear
    Spdx2_RelationshipType.OPTIONAL_DEPENDENCY_OF: (
        SoftwareDependencyRelationship,
        RelationshipType.DEPENDS_ON,
        {"conditionality": DependencyConditionalityType.OPTIONAL},
    ),
    Spdx2_RelationshipType.OTHER: (Relationship, RelationshipType.OTHER, {}),
    Spdx2_RelationshipType.PACKAGE_OF: (SoftwareDependencyRelationship, RelationshipType.DEPENDS_ON, {}),
    Spdx2_RelationshipType.PATCH_APPLIED: (Relationship, RelationshipType.PATCH, {"swap": True}),
    Spdx2_RelationshipType.PATCH_FOR: (Relationship, RelationshipType.PATCH, {}),
    Spdx2_RelationshipType.PREREQUISITE_FOR: (
        SoftwareDependencyRelationship,
        RelationshipType.DEPENDS_ON,
        {"conditionality": DependencyConditionalityType.PREREQUISITE},
    ),
    Spdx2_RelationshipType.PROVIDED_DEPENDENCY_OF: (
        SoftwareDependencyRelationship,
        RelationshipType.DEPENDS_ON,
        {"scope": LifecycleScopeType.BUILD, "conditionality": DependencyConditionalityType.PROVIDED},
    ),
    Spdx2_RelationshipType.RUNTIME_DEPENDENCY_OF: (
        SoftwareDependencyRelationship,
        RelationshipType.DEPENDS_ON,
        {"scope": LifecycleScopeType.RUNTIME},
    ),
    Spdx2_RelationshipType.STATIC_LINK: (
        SoftwareDependencyRelationship,
        RelationshipType.DEPENDS_ON,
        {"linkage": SoftwareDependencyLinkType.STATIC},
    ),
    Spdx2_RelationshipType.TEST_CASE_OF: (Relationship, RelationshipType.TEST_CASE, {}),
    Spdx2_RelationshipType.TEST_DEPENDENCY_OF: (
        SoftwareDependencyRelationship,
        RelationshipType.DEPENDS_ON,
        {"scope": LifecycleScopeType.TEST},
    ),
    Spdx2_RelationshipType.TEST_OF: (Relationship, RelationshipType.TEST, {}),
    Spdx2_RelationshipType.TEST_TOOL_OF: (
        SoftwareDependencyRelationship,
        RelationshipType.DEPENDS_ON,
        {"scope": LifecycleScopeType.TEST, "linkage": SoftwareDependencyLinkType.TOOL},
    ),
    Spdx2_RelationshipType.VARIANT_OF: (Relationship, RelationshipType.VARIANT, {}),
    Spdx2_RelationshipType.REQUIREMENT_DESCRIPTION_FOR: (Relationship, RelationshipType.REQUIREMENT_FOR, {}),
    Spdx2_RelationshipType.SPECIFICATION_FOR: (Relationship, RelationshipType.SPECIFICATION_FOR, {}),
}


def bump_relationships(
    spdx2_relationships: List[Spdx2_Relationship],
    payload: Payload,
    document_namespace: str,
):
    generated_relationships: Dict[Tuple[str, str], List[Relationship]] = {}
    for counter, spdx2_relationship in enumerate(spdx2_relationships):
        relationship = bump_relationship(spdx2_relationship, document_namespace, counter)
        if relationship:
            generated_relationships.setdefault(
                (relationship.from_element, relationship.relationship_type.name), []
            ).append(relationship)

    for relationships in generated_relationships.values():
        if len(relationships) > 1:
            _merge_relationships_and_add_to_payload(relationships, payload)
        else:
            payload.add_element(relationships[0])


def bump_relationship(
    spdx2_relationship: Spdx2_Relationship,
    document_namespace: str,
    counter: int,
) -> Optional[Union[Relationship, SoftwareDependencyRelationship]]:
    completeness, to = determine_completeness_and_to(spdx2_relationship.related_spdx_element_id)
    spdx_id = "#".join([document_namespace, f"SPDXRef-Relationship-{counter}"])
    relationship_class, relationship_type, parameters = relationship_mapping[spdx2_relationship.relationship_type]
    if relationship_class is None:
        print_missing_conversion(spdx2_relationship.relationship_type.name, 0)
        return

    swap_direction = parameters.get("swap", False)

    if swap_direction:
        if not to:
            print_missing_conversion("Swapped Relationship to NoAssertion/None", 0)
            return
        from_element = to[0]
        to = [spdx2_relationship.spdx_element_id]
    else:
        from_element = spdx2_relationship.spdx_element_id

    if relationship_class == SoftwareDependencyRelationship:
        from_element = spdx2_relationship.spdx_element_id

        return SoftwareDependencyRelationship(
            spdx_id,
            f"{document_namespace}#{from_element}",
            relationship_type,
            [f"{document_namespace}#{t}" for t in to],
            comment=spdx2_relationship.comment,
            completeness=completeness,
            scope=parameters.get("scope"),
            software_linkage=parameters.get("linkage"),
            conditionality=parameters.get("conditionality"),
        )

    return Relationship(
        spdx_id,
        f"{document_namespace}#{from_element}",
        relationship_type,
        [f"{document_namespace}#{t}" for t in to],
        comment=spdx2_relationship.comment,
        completeness=completeness,
    )


def determine_completeness_and_to(
    related_spdx_element_id: Union[str, SpdxNone, SpdxNoAssertion]
) -> Tuple[Optional[RelationshipCompleteness], List[str]]:
    if isinstance(related_spdx_element_id, SpdxNoAssertion):
        completeness = RelationshipCompleteness.NOASSERTION
        to = []
    elif isinstance(related_spdx_element_id, SpdxNone):
        completeness = RelationshipCompleteness.COMPLETE
        to = []
    else:
        completeness = None
        to = [related_spdx_element_id]
    return completeness, to


def _merge_relationships_and_add_to_payload(relationships: List[Relationship], payload: Payload):
    to = []
    completeness = None
    spdx_id = None
    merged_relationship = relationships[0]
    for merged_relationship in relationships:
        if merged_relationship.comment:
            payload.add_element(merged_relationship)
            continue
        if merged_relationship.completeness:
            if completeness and completeness != merged_relationship.completeness:
                logging.warning(
                    f"Contradicting information about completeness of relationship: {merged_relationship}", sys.stderr
                )
            else:
                completeness = merged_relationship.completeness

        to += merged_relationship.to
        spdx_id = merged_relationship.spdx_id
    if to:
        merged_relationship.spdx_id = spdx_id
        merged_relationship.to = to
        merged_relationship.completeness = completeness
        merged_relationship.comment = None
        payload.add_element(merged_relationship)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import Enum, auto

from spdx_tools.spdx.parser.error import SPDXParsingError


class FileFormat(Enum):
    JSON = auto()
    YAML = auto()
    XML = auto()
    TAG_VALUE = auto()
    RDF_XML = auto()


def file_name_to_format(file_name: str) -> FileFormat:
    if file_name.endswith(".rdf") or file_name.endswith(".rdf.xml"):
        return FileFormat.RDF_XML
    elif file_name.endswith(".tag") or file_name.endswith(".spdx"):
        return FileFormat.TAG_VALUE
    elif file_name.endswith(".json"):
        return FileFormat.JSON
    elif file_name.endswith(".xml"):
        return FileFormat.XML
    elif file_name.endswith(".yaml") or file_name.endswith(".yml"):
        return FileFormat.YAML
    else:
        raise SPDXParsingError(["Unsupported SPDX file type: " + str(file_name)])



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import field
from enum import Enum, auto

from beartype.typing import List, Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values


class ExternalReferenceType(Enum):
    ALT_DOWNLOAD_LOCATION = auto()
    ALT_WEB_PAGE = auto()
    BINARY_ARTIFACT = auto()
    BUILD_META = auto()
    BUILD_SYSTEM = auto()
    CERTIFICATION_REPORT = auto()
    CHAT = auto()
    COMPONENT_ANALYSIS_REPORT = auto()
    DOCUMENTATION = auto()
    DYNAMIC_ANALYSIS_REPORT = auto()
    EOL_NOTICE = auto()
    FUNDING = auto()
    ISSUE_TRACKER = auto()
    LICENSE = auto()
    MAILING_LIST = auto()
    METRICS = auto()
    OTHER = auto()
    PRODUCT_METADATA = auto()
    QUALITY_ASSESSMENT_REPORT = auto()
    RELEASE_HISTORY = auto()
    RELEASE_NOTES = auto()
    RISK_ASSESSMENT = auto()
    RUNTIME_ANALYSIS_REPORT = auto()
    SECURE_SOFTWARE_ATTESTATION = auto()
    SECURITY_ADVERSARY_MODEL = auto()
    SECURITY_ADVISORY = auto()
    SECURITY_FIX = auto()
    SECURITY_OTHER = auto()
    SECURITY_PEN_TEST_REPORT = auto()
    SECURITY_POLICY = auto()
    SECURITY_THREAT_MODEL = auto()
    SOCIAL_MEDIA = auto()
    SOURCE_ARTIFACT = auto()
    STATIC_ANALYSIS_REPORT = auto()
    SUPPORT = auto()
    VCS = auto()
    VULNERABILITY_DISCLOSURE_REPORT = auto()
    VULNERABILITY_EXPLOITABILITY_ASSESSMENT = auto()


@dataclass_with_properties
class ExternalReference:
    external_reference_type: Optional[ExternalReferenceType] = None
    locator: List[str] = field(default_factory=list)
    content_type: Optional[str] = None  # placeholder for MediaType
    comment: Optional[str] = None

    def __init__(
        self,
        external_reference_type: Optional[ExternalReferenceType] = None,
        locator: List[str] = None,
        content_type: Optional[str] = None,
        comment: Optional[str] = None,
    ):
        locator = [] if locator is None else locator
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from rdflib import BNode, Graph

from spdx_tools.spdx.model import Checksum, ChecksumAlgorithm
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import (
    construct_or_raise_parsing_error,
    raise_parsing_error_if_logger_has_messages,
)
from spdx_tools.spdx.parser.rdf.graph_parsing_functions import parse_literal, remove_prefix
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE


def parse_checksum(parent_node: BNode, graph: Graph) -> Checksum:
    logger = Logger()
    algorithm = parse_literal(
        logger, graph, parent_node, SPDX_NAMESPACE.algorithm, parsing_method=convert_rdf_to_algorithm
    )
    value = parse_literal(logger, graph, parent_node, SPDX_NAMESPACE.checksumValue)

    raise_parsing_error_if_logger_has_messages(logger, "Checksum")
    checksum = construct_or_raise_parsing_error(Checksum, dict(algorithm=algorithm, value=value))
    return checksum


def convert_rdf_to_algorithm(algorithm: str) -> ChecksumAlgorithm:
    algorithm = remove_prefix(algorithm, SPDX_NAMESPACE.checksumAlgorithm_).upper()
    if "BLAKE2B" in algorithm:
        algorithm = algorithm.replace("BLAKE2B", "BLAKE2B_")
    try:
        checksum = ChecksumAlgorithm[algorithm]
    except KeyError:
        raise SPDXParsingError([f"Invalid value for ChecksumAlgorithm: {algorithm}"])
    return checksum



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import field

from beartype.typing import List, Optional, Tuple, Union
from license_expression import LicenseExpression

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx.model import SpdxNoAssertion, SpdxNone


@dataclass_with_properties
class Snippet:
    spdx_id: str
    file_spdx_id: str
    byte_range: Tuple[int, int]
    line_range: Optional[Tuple[int, int]] = None
    license_concluded: Optional[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]] = None
    license_info_in_snippet: List[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]] = None
    license_comment: Optional[str] = None
    copyright_text: Optional[Union[str, SpdxNoAssertion, SpdxNone]] = None
    comment: Optional[str] = None
    name: Optional[str] = None
    attribution_texts: List[str] = field(default_factory=list)

    def __init__(
        self,
        spdx_id: str,
        file_spdx_id: str,
        byte_range: Tuple[int, int],
        line_range: Optional[Tuple[int, int]] = None,
        license_concluded: Optional[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]] = None,
        license_info_in_snippet: List[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]] = None,
        license_comment: Optional[str] = None,
        copyright_text: Optional[Union[str, SpdxNoAssertion, SpdxNone]] = None,
        comment: Optional[str] = None,
        name: Optional[str] = None,
        attribution_texts: List[str] = None,
    ):
        attribution_texts = [] if attribution_texts is None else attribution_texts
        license_info_in_snippet = [] if license_info_in_snippet is None else license_info_in_snippet
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx.jsonschema.external_package_ref_converter import ExternalPackageRefConverter
from spdx_tools.spdx.jsonschema.external_package_ref_properties import ExternalPackageRefProperty
from spdx_tools.spdx.model import ExternalPackageRef, ExternalPackageRefCategory


@pytest.fixture
def converter() -> ExternalPackageRefConverter:
    return ExternalPackageRefConverter()


@pytest.mark.parametrize(
    "external_package_ref_property,expected",
    [
        (ExternalPackageRefProperty.COMMENT, "comment"),
        (ExternalPackageRefProperty.REFERENCE_CATEGORY, "referenceCategory"),
        (ExternalPackageRefProperty.REFERENCE_LOCATOR, "referenceLocator"),
        (ExternalPackageRefProperty.REFERENCE_TYPE, "referenceType"),
    ],
)
def test_json_property_names(
    converter: ExternalPackageRefConverter, external_package_ref_property: ExternalPackageRefProperty, expected: str
):
    assert converter.json_property_name(external_package_ref_property) == expected


def test_json_type(converter: ExternalPackageRefConverter):
    assert converter.get_json_type() == ExternalPackageRefProperty


def test_data_model_type(converter: ExternalPackageRefConverter):
    assert converter.get_data_model_type() == ExternalPackageRef


def test_successful_conversion(converter: ExternalPackageRefConverter):
    external_package_ref = ExternalPackageRef(ExternalPackageRefCategory.PACKAGE_MANAGER, "type", "locator", "comment")

    converted_dict = converter.convert(external_package_ref)

    assert converted_dict == {
        converter.json_property_name(ExternalPackageRefProperty.COMMENT): "comment",
        converter.json_property_name(ExternalPackageRefProperty.REFERENCE_CATEGORY): "PACKAGE_MANAGER",
        converter.json_property_name(ExternalPackageRefProperty.REFERENCE_LOCATOR): "locator",
        converter.json_property_name(ExternalPackageRefProperty.REFERENCE_TYPE): "type",
    }



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import re

from beartype.typing import List

from spdx_tools.spdx.model import PackageVerificationCode
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage


def validate_verification_code(verification_code: PackageVerificationCode, parent_id: str) -> List[ValidationMessage]:
    validation_messages: List[ValidationMessage] = []
    context = ValidationContext(
        parent_id=parent_id, element_type=SpdxElementType.PACKAGE_VERIFICATION_CODE, full_element=verification_code
    )

    for file in verification_code.excluded_files:
        if file.startswith("/"):
            validation_messages.append(
                ValidationMessage(f'file name must not be an absolute path starting with "/", but is: {file}', context)
            )

    value: str = verification_code.value
    if not re.match("^[0-9a-f]{40}$", value):
        validation_messages.append(
            ValidationMessage(
                f"value of verification_code must consist of 40 lowercase hexadecimal digits, but is: {value} "
                f"(length: {len(value)} digits)",
                context,
            )
        )

    return validation_messages



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest
from license_expression import get_spdx_licensing

from spdx_tools.spdx.model import FileType, SpdxNoAssertion
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.tagvalue.parser import Parser
from tests.spdx.parser.tagvalue.test_creation_info_parser import DOCUMENT_STR


def test_parse_file():
    parser = Parser()
    file_str = "\n".join(
        [
            "FileName: testfile.java",
            "SPDXID: SPDXRef-File",
            "FileType: SOURCE",
            "FileType: TEXT",
            "FileChecksum: SHA1: 2fd4e1c67a2d28fced849ee1bb76e7391b93eb12",
            "LicenseConcluded: Apache-2.0",
            "LicenseInfoInFile: Apache-2.0",
            "LicenseInfoInFile: NOASSERTION",
            "FileCopyrightText: <text>Copyright 2014 Acme Inc.</text>",
            "FileComment: <text>Very long file</text>",
            "FileAttributionText: <text>Acknowledgements that might be required to be communicated in some contexts."
            "</text>",
        ]
    )
    document = parser.parse("\n".join([DOCUMENT_STR, file_str]))
    assert document is not None
    assert len(document.files) == 1
    spdx_file = document.files[0]
    assert spdx_file.name == "testfile.java"
    assert spdx_file.spdx_id == "SPDXRef-File"
    assert spdx_file.file_types == [FileType.SOURCE, FileType.TEXT]
    assert spdx_file.comment == "Very long file"
    assert spdx_file.attribution_texts == [
        "Acknowledgements that might be required to be communicated in some contexts."
    ]
    assert spdx_file.license_info_in_file == [get_spdx_licensing().parse("Apache-2.0"), SpdxNoAssertion()]
    assert spdx_file.license_concluded == get_spdx_licensing().parse("Apache-2.0")


def test_parse_invalid_file():
    parser = Parser()
    file_str = "\n".join(
        [
            "FileName: testfile.java",
            "SPDXID: SPDXRef-File",
            "FileType: SOUCE",
            "FileType: TEXT",
            "FileChecksum: SHA3: 2fd4e1c67a2d28fced849ee1bb76e7391b93eb12",
            "LicenseConcluded: Apache-2.0",
            "LicenseInfoInFile: Apache-2.0",
            "FileCopyrightText: <text>Copyright 2014 Acme Inc.</text>",
            "FileComment: <text>Very long file</text>",
            "FileAttributionText: <text>Acknowledgements that might be required to be communicated in some contexts."
            "</text>",
        ]
    )

    with pytest.raises(SPDXParsingError) as err:
        parser.parse(file_str)

    assert err.value.get_messages() == [
        "Error while parsing File: ['Invalid FileType: SOUCE. Line 3', 'Error while "
        "parsing FileChecksum: Token did not match specified grammar rule. Line: 5']"
    ]



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime

import pytest

from spdx_tools.spdx.datetime_conversions import datetime_to_iso_string
from spdx_tools.spdx.jsonschema.creation_info_converter import CreationInfoConverter
from spdx_tools.spdx.jsonschema.creation_info_properties import CreationInfoProperty
from spdx_tools.spdx.model import Actor, ActorType, CreationInfo, Version
from tests.spdx.fixtures import creation_info_fixture


@pytest.fixture
def converter() -> CreationInfoConverter:
    return CreationInfoConverter()


@pytest.mark.parametrize(
    "creation_info_property,expected",
    [
        (CreationInfoProperty.CREATED, "created"),
        (CreationInfoProperty.CREATORS, "creators"),
        (CreationInfoProperty.LICENSE_LIST_VERSION, "licenseListVersion"),
        (CreationInfoProperty.COMMENT, "comment"),
    ],
)
def test_json_property_names(
    converter: CreationInfoConverter, creation_info_property: CreationInfoProperty, expected: str
):
    assert converter.json_property_name(creation_info_property) == expected


def test_successful_conversion(converter: CreationInfoConverter):
    creators = [Actor(ActorType.PERSON, "personName"), Actor(ActorType.TOOL, "toolName")]
    created = datetime(2022, 12, 1)

    converted_dict = converter.convert(
        creation_info_fixture(
            creators=creators, created=created, creator_comment="comment", license_list_version=Version(1, 2)
        )
    )

    assert converted_dict == {
        converter.json_property_name(CreationInfoProperty.CREATED): datetime_to_iso_string(created),
        converter.json_property_name(CreationInfoProperty.CREATORS): ["Person: personName", "Tool: toolName"],
        converter.json_property_name(CreationInfoProperty.LICENSE_LIST_VERSION): "1.2",
        converter.json_property_name(CreationInfoProperty.COMMENT): "comment",
    }


def test_null_values(converter: CreationInfoConverter):
    creation_info = creation_info_fixture(license_list_version=None, creator_comment=None, creators=[])

    converted_dict = converter.convert(creation_info)

    assert converter.json_property_name(CreationInfoProperty.LICENSE_LIST_VERSION) not in converted_dict
    assert converter.json_property_name(CreationInfoProperty.COMMENT) not in converted_dict
    assert converter.json_property_name(CreationInfoProperty.CREATORS) not in converted_dict


def test_json_type(converter: CreationInfoConverter):
    assert converter.get_json_type() == CreationInfoProperty


def test_data_model_type(converter: CreationInfoConverter):
    assert converter.get_data_model_type() == CreationInfo



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from unittest import TestCase

import pytest

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.model import Actor, ActorType, Checksum, ChecksumAlgorithm, ExternalDocumentRef, Version
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.tagvalue.parser import Parser

DOCUMENT_STR = "\n".join(
    [
        "SPDXVersion: SPDX-2.3",
        "DataLicense: CC0-1.0",
        "DocumentName: Sample_Document-V2.3",
        f"SPDXID: {DOCUMENT_SPDX_ID}",
        "DocumentComment: <text>Sample Comment</text>",
        "DocumentNamespace: https://spdx.org/spdxdocs/spdx-example-444504E0-4F89-41D3-9A0C-0305E82C3301",
        "ExternalDocumentRef: DocumentRef-spdx-tool-1.2 "
        "http://spdx.org/spdxdocs/spdx-tools-v1.2-3F2504E0-4F89-41D3-9A0C-0305E82C3301 "
        "SHA1: d6a770ba38583ed4bb4525bd96e50461655d2759",
        "Creator: Person: Bob (bob@example.com)",
        "Creator: Organization: Acme.",
        "Created: 2010-02-03T00:00:00Z",
        "CreatorComment: <text>Sample Comment \nwith multiple \nlines.</text>",
        "LicenseListVersion: 3.17",
    ]
)


def test_parse_creation_info():
    parser = Parser()
    document = parser.parse(DOCUMENT_STR)
    assert document is not None
    creation_info = document.creation_info
    assert creation_info is not None
    assert creation_info.spdx_version == "SPDX-2.3"
    assert creation_info.data_license == "CC0-1.0"
    assert creation_info.name == "Sample_Document-V2.3"
    assert creation_info.spdx_id == DOCUMENT_SPDX_ID
    assert creation_info.document_comment == "Sample Comment"
    assert (
        creation_info.document_namespace
        == "https://spdx.org/spdxdocs/spdx-example-444504E0-4F89-41D3-9A0C-0305E82C3301"
    )
    TestCase().assertCountEqual(
        creation_info.creators,
        [Actor(ActorType.PERSON, "Bob", "bob@example.com"), Actor(ActorType.ORGANIZATION, "Acme.")],
    )
    assert creation_info.creator_comment == "Sample Comment \nwith multiple \nlines."
    assert creation_info.created == datetime(2010, 2, 3)
    assert creation_info.license_list_version == Version(3, 17)
    assert creation_info.external_document_refs == [
        ExternalDocumentRef(
            "DocumentRef-spdx-tool-1.2",
            "http://spdx.org/spdxdocs/spdx-tools-v1.2-3F2504E0-4F89-41D3-9A0C-0305E82C3301",
            Checksum(ChecksumAlgorithm.SHA1, "d6a770ba38583ed4bb4525bd96e50461655d2759"),
        )
    ]


@pytest.mark.parametrize(
    "document_str, expected_message",
    (
        [
            (
                "\n".join(
                    [
                        "SPDXVersion: SPDX-2.3",
                        "DataLicense: CC0-1.0",
                        "DocumentName: Sample_Document-V2.3",
                        f"SPDXID: {DOCUMENT_SPDX_ID}",
                        "DocumentComment: <text>Sample Comment</text>",
                        "DocumentNamespace: <text>Sample Comment</text>",
                        "ExternalDocumentRef: DocumentRef-spdx-tool-1.2:htp://spdx.org:SHA1: "
                        "d6a770ba38583ed4bb4525bd96e50461655d2759",
                        "Creator: Person Bob (bob@example.com)",
                        "Creator: Organization: Acme [email]",
                        "Created: 2010-02-03T00:00:0Z",
                        "CreatorComment: <text>Sample Comment</text>",
                        "LicenseListVersion: 7",
                    ]
                ),
                (
                    "Error while parsing CreationInfo: ['Error while parsing DocumentNamespace: "
                    "Token did not match specified grammar rule. Line: 6', \"Error while parsing "
                    "ExternalDocumentRef: Couldn't split the first part of the value into "
                    "document_ref_id and document_uri. Line: 7\", 'Error while parsing Creator: "
                    "Token did not match specified grammar rule. Line: 8', 'Error while parsing "
                    "Created: Token did not match specified grammar rule. Line: 10', '7 is not a "
                    "valid version string']"
                ),
            ),
            (
                "\n".join(
                    [
                        "SPDXVersion: SPDX-2.3",
                        "DataLicense: CC0-1.0",
                        "DocumentName: Sample_Document-V2.3",
                        f"SPDXID: {DOCUMENT_SPDX_ID}",
                    ]
                ),
                r"__init__() missing 3 required positional arguments: 'document_namespace', 'creators', and 'created'",
            ),
            (
                "LicenseListVersion: 3.5\nLicenseListVersion: 3.7",
                "Error while parsing CreationInfo: ['Multiple values for LicenseListVersion found. Line: 2']",
            ),
            (
                "ExternalDocumentRef: Document_ref document_uri SHA1: afded",
                "Error while parsing CreationInfo: [\"Error while parsing ExternalDocumentRef: Couldn't match "
                'Checksum. Line: 1"]',
            ),
        ]
    ),
)
def test_parse_invalid_creation_info(document_str, expected_message):
    parser = Parser()
    with pytest.raises(SPDXParsingError) as err:
        parser.parse(document_str)
    assert expected_message in err.value.get_messages()[0]



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from abc import ABC, abstractmethod
from dataclasses import field

from beartype.typing import List, Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties


@dataclass_with_properties
class LicenseAddition(ABC):
    addition_id: str
    addition_name: str
    addition_text: str
    addition_comment: Optional[str] = None
    see_also: List[str] = field(default_factory=list)
    standard_addition_template: Optional[str] = None
    is_deprecated_addition_id: Optional[bool] = None
    obsoleted_by: Optional[str] = None

    @abstractmethod
    def __init__(self):
        pass



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import Enum


class JsonProperty(Enum):
    """
    Parent class for all json property classes. Not meant to be instantiated directly, only to have a common parent
    type that can be used in type hints.
    In general, all the child enums list the properties of the corresponding objects from the json schema.
    """

    pass
